import pytest

import toga
from toga.constants import FlashMode
from toga.hardware.camera import CameraDevice
from toga_dummy import factory
from toga_dummy.hardware.camera import (
    Camera as DummyCamera,
    CameraDevice as DummyCameraDevice,
)
from toga_dummy.utils import (
    assert_action_not_performed,
    assert_action_performed,
    assert_action_performed_with,
)


@pytest.fixture
def photo(app):
    return toga.Image("resources/photo.png")


def test_no_camera(monkeypatch, app):
    """If there's no camera, and no factory implementation, accessing camera raises an
    exception."""
    try:
        monkeypatch.delattr(app, "_camera")
    except AttributeError:
        pass
    monkeypatch.delattr(factory, "Camera")

    # Accessing the camera object should raise NotImplementedError
    with pytest.raises(NotImplementedError):
        app.camera


@pytest.mark.parametrize(
    "initial, should_request, has_permission",
    [
        (-1, True, True),
        (0, True, False),
        (1, False, True),
    ],
)
def test_request_permission(app, initial, should_request, has_permission):
    """An app can request permission to use the camera."""
    # The camera instance round-trips the app instance
    assert app.camera.app == app

    # Set initial permission
    app.camera._impl._has_permission = initial

    assert (
        app.loop.run_until_complete(app.camera.request_permission()) == has_permission
    )

    if should_request:
        assert_action_performed(app.camera, "request permission")
    else:
        assert_action_not_performed(app.camera, "request permission")

    # As a result of requesting, photo permission is as expected
    assert app.camera.has_permission == has_permission


def test_request_permission_sync(app):
    """An app can synchronously request permission to use the camera."""
    # Set initial permission
    app.camera._impl._has_permission = -1

    result = app.camera.request_permission()

    # This will cause a permission request to occur...
    assert_action_performed(app.camera, "request permission")

    # ... but the result won't be directly comparable
    with pytest.raises(RuntimeError):
        # == True isn't good python, but it's going to raise an exception anyway.
        result == True  # noqa: E712


def test_device_properties(app):
    """Device properties can be checked."""

    assert [
        {
            "device": device,
            "__repr__": repr(device),
            "__str__": str(device),
            "name": device.name,
            "id": device.id,
            "has_flash": device.has_flash,
        }
        for device in app.camera.devices
    ] == [
        {
            "device": CameraDevice(DummyCamera.CAMERA_1),
            "__repr__": "<CameraDevice id=camera-1 'Camera 1'>",
            "__str__": "Camera 1",
            "name": "Camera 1",
            "id": "camera-1",
            "has_flash": True,
        },
        {
            "device": CameraDevice(DummyCamera.CAMERA_2),
            "__repr__": "<CameraDevice id=camera-2 'Camera 2'>",
            "__str__": "Camera 2",
            "name": "Camera 2",
            "id": "camera-2",
            "has_flash": False,
        },
    ]

    # Identity check
    assert CameraDevice(DummyCamera.CAMERA_1) == CameraDevice(DummyCamera.CAMERA_1)
    # A different instance with the same ID is equal
    duplicate = CameraDevice(
        DummyCameraDevice(id="camera-1", name="Duplicate Camera 1", has_flash=True)
    )
    assert CameraDevice(DummyCamera.CAMERA_1) == duplicate
    # Different cameras aren't equal
    assert CameraDevice(DummyCamera.CAMERA_1) != CameraDevice(DummyCamera.CAMERA_2)


@pytest.mark.parametrize(
    "device",
    [None, CameraDevice(DummyCamera.CAMERA_1), CameraDevice(DummyCamera.CAMERA_2)],
)
@pytest.mark.parametrize(
    "flash",
    [FlashMode.AUTO, FlashMode.ON, FlashMode.OFF],
)
def test_take_photo_with_permission(app, device, flash, photo):
    """If permission has not been previously requested, it is requested before a photo
    is taken."""
    # Set permission to potentially allowed
    app.camera._impl._has_permission = -1

    app.camera._impl.simulate_photo(photo)

    result = app.loop.run_until_complete(
        app.camera.take_photo(device=device, flash=flash)
    )

    # Photo was returned
    assert result == photo

    assert_action_performed(app.camera, "has permission")
    assert_action_performed_with(
        app.camera,
        "take photo",
        permission_requested=True,
        device=device,
        flash=flash,
    )


def test_take_photo_prior_permission(app, photo):
    """If permission has been previously requested, a photo can be taken."""
    # Set permission
    app.camera._impl._has_permission = 1

    # Simulate the camera response
    app.camera._impl.simulate_photo(photo)

    result = app.loop.run_until_complete(app.camera.take_photo())

    # Photo was returned
    assert result == photo

    assert_action_performed(app.camera, "has permission")
    assert_action_performed_with(
        app.camera,
        "take photo",
        permission_requested=False,
        device=None,
        flash=FlashMode.AUTO,
    )


def test_take_photo_no_permission(app, photo):
    """If permission has been denied, an exception is raised."""
    # Deny permission
    app.camera._impl._has_permission = 0

    with pytest.raises(
        PermissionError,
        match=r"App does not have permission to take photos",
    ):
        app.loop.run_until_complete(app.camera.take_photo())

    assert_action_performed(app.camera, "has permission")
    assert_action_not_performed(app.camera, "take photo")



import pytest

import toga
from toga_dummy.utils import (
    EventLog,
    assert_action_not_performed,
    assert_action_performed,
    attribute_value,
)


@pytest.fixture
def label():
    return toga.Label("Test Label")


def test_label_created(label):
    """A label can be created."""
    # Round trip the impl/interface
    assert label._impl.interface == label
    assert_action_performed(label, "create Label")


@pytest.mark.parametrize(
    "value, expected",
    [
        ("New Text", "New Text"),
        (12345, "12345"),
        (None, ""),
        ("\u200B", ""),
        ("Contains\nsome\nnewlines", "Contains\nsome\nnewlines"),
    ],
)
def test_update_label_text(label, value, expected):
    assert label.text == "Test Label"

    # Clear the event log
    EventLog.reset()

    label.text = value
    assert label.text == expected

    # test backend has the right value
    assert attribute_value(label, "text") == expected

    # A rehint was performed
    assert_action_performed(label, "refresh")


def test_focus_noop(label):
    """Focus is a no-op."""

    label.focus()
    assert_action_not_performed(label, "focus")



from __future__ import annotations

from typing import NamedTuple


class LatLng(NamedTuple):
    """A geographic coordinate."""

    #: Latitude
    lat: float

    #: Longitude
    lng: float

    def __str__(self):
        return f"({self.lat:6f}, {self.lng:6f})"



from pathlib import Path

import pytest

import toga
from toga.fonts import (
    _REGISTERED_FONT_CACHE,
    BOLD,
    ITALIC,
    NORMAL,
    SANS_SERIF,
    SMALL_CAPS,
    SYSTEM,
    SYSTEM_DEFAULT_FONT_SIZE,
)


@pytest.fixture
def app():
    return toga.App("Fonts Test", "org.beeware.toga.fonts")


@pytest.mark.parametrize(
    "family, size, weight, style, variant, as_str",
    [
        # No modifiers
        (
            SANS_SERIF,
            12,
            NORMAL,
            NORMAL,
            NORMAL,
            "sans-serif 12pt",
        ),
        # Weight modifier
        (
            SANS_SERIF,
            13,
            BOLD,
            NORMAL,
            NORMAL,
            "sans-serif 13pt bold",
        ),
        # Style modifier
        (
            SANS_SERIF,
            14,
            NORMAL,
            ITALIC,
            NORMAL,
            "sans-serif 14pt italic",
        ),
        # Variant modifier
        (
            SANS_SERIF,
            15,
            NORMAL,
            NORMAL,
            SMALL_CAPS,
            "sans-serif 15pt small-caps",
        ),
        # All modifiers
        (
            SANS_SERIF,
            37,
            BOLD,
            ITALIC,
            SMALL_CAPS,
            "sans-serif 37pt bold small-caps italic",
        ),
        # System font, fixed size
        (
            SYSTEM,
            42,
            NORMAL,
            NORMAL,
            NORMAL,
            "system 42pt",
        ),
        # Custom font, default size
        (
            "Custom Font",
            SYSTEM_DEFAULT_FONT_SIZE,
            NORMAL,
            NORMAL,
            NORMAL,
            "Custom Font default size",
        ),
        # System font, default size
        (
            SYSTEM,
            SYSTEM_DEFAULT_FONT_SIZE,
            NORMAL,
            NORMAL,
            NORMAL,
            "system default size",
        ),
    ],
)
def test_builtin_font(family, size, weight, style, variant, as_str):
    """A builtin font can be constructed."""
    font = toga.Font(
        family=family,
        size=size,
        style=style,
        weight=weight,
        variant=variant,
    )

    assert font.family == family
    assert font.size == size
    assert font.style == style
    assert font.weight == weight
    assert font.variant == variant
    assert str(font) == as_str


@pytest.mark.parametrize(
    "family, weight, style, variant, key",
    [
        ("Helvetica", NORMAL, NORMAL, NORMAL, ("Helvetica", NORMAL, NORMAL, NORMAL)),
        (
            "Times New Roman",
            BOLD,
            ITALIC,
            SMALL_CAPS,
            ("Times New Roman", BOLD, ITALIC, SMALL_CAPS),
        ),
        # Unknown style/weight/variants are normalized to "NORMAL"
        ("Wonky", "unknown", ITALIC, SMALL_CAPS, ("Wonky", NORMAL, ITALIC, SMALL_CAPS)),
        ("Wonky", BOLD, "unknown", SMALL_CAPS, ("Wonky", BOLD, NORMAL, SMALL_CAPS)),
        ("Wonky", BOLD, ITALIC, "unknown", ("Wonky", BOLD, ITALIC, NORMAL)),
    ],
)
def test_registered_font_key(app, family, style, weight, variant, key):
    """Registered font keys can be generated."""
    assert (
        toga.Font._registered_font_key(
            family, style=style, weight=weight, variant=variant
        )
        == key
    )


@pytest.mark.parametrize(
    "path, registered",
    [
        # Absolute path
        (Path("/path/to/custom/font.otf"), Path("/path/to/custom/font.otf")),
        (str(Path("/path/to/custom/font.otf")), Path("/path/to/custom/font.otf")),
        # Relative path
        (
            Path("path/to/custom/font.otf"),
            Path(toga.__file__).parent / "path/to/custom/font.otf",
        ),
        (
            "path/to/custom/font.otf",
            Path(toga.__file__).parent / "path/to/custom/font.otf",
        ),
    ],
)
def test_register_font(app, path, registered):
    """A custom font can be registered."""
    toga.Font.register("Custom Font", path)

    # Test fixture has paths in Path format; fully resolve for test comparison. This
    # gets around Windows path separator and absolute path discrepancies.
    assert (
        Path(_REGISTERED_FONT_CACHE[("Custom Font", NORMAL, NORMAL, NORMAL)]).resolve()
        == registered.resolve()
    )


@pytest.mark.parametrize(
    "path, registered",
    [
        # Absolute path
        (Path("/path/to/custom/font.otf"), Path("/path/to/custom/font.otf")),
        (str(Path("/path/to/custom/font.otf")), Path("/path/to/custom/font.otf")),
        # Relative path
        (
            Path("path/to/custom/font.otf"),
            Path(toga.__file__).parent / "path/to/custom/font.otf",
        ),
        (
            str(Path("path/to/custom/font.otf")),
            Path(toga.__file__).parent / "path/to/custom/font.otf",
        ),
    ],
)
def test_register_font_variant(app, path, registered):
    """A custom font can be registered as a variant."""
    toga.Font.register("Custom Font", path, weight=BOLD)

    # Test fixture has paths in Path format; fully resolve for test comparison. This
    # gets around Windows path separator and absolute path discrepancies.
    assert (
        Path(_REGISTERED_FONT_CACHE[("Custom Font", BOLD, NORMAL, NORMAL)]).resolve()
        == registered.resolve()
    )



from pathlib import Path
from unittest.mock import Mock

import pytest

import toga
from toga_dummy.utils import (
    assert_action_not_performed,
    assert_action_performed,
    assert_action_performed_with,
)


@pytest.fixture
def window(app):
    return toga.Window()


def test_window_created(app):
    """A Window can be created with minimal arguments."""
    window = toga.Window()

    assert window.app == app
    assert window.content is None

    assert window._impl.interface == window
    assert_action_performed(window, "create Window")

    # We can't know what the ID is, but it must be a string.
    assert isinstance(window.id, str)
    assert window.title == "Toga"
    assert window.position == (100, 100)
    assert window.size == (640, 480)
    assert window.resizable
    assert window.closable
    assert window.minimizable
    assert len(window.toolbar) == 0
    assert window.on_close._raw is None


def test_window_created_explicit(app):
    """Explicit arguments at construction are stored."""
    on_close_handler = Mock()

    window = toga.Window(
        id="my-window",
        title="My Window",
        position=(10, 20),
        size=(200, 300),
        resizable=False,
        closable=False,
        minimizable=False,
        on_close=on_close_handler,
    )

    assert window.app == app
    assert window.content is None

    assert window._impl.interface == window
    assert_action_performed(window, "create Window")

    assert window.id == "my-window"
    assert window.title == "My Window"
    assert window.position == (10, 20)
    assert window.size == (200, 300)
    assert not window.resizable
    assert not window.closable
    assert not window.minimizable
    assert len(window.toolbar) == 0
    assert window.on_close._raw == on_close_handler


def test_window_created_without_app():
    """A window cannot be created without an active app."""
    toga.App.app = None
    with pytest.raises(
        RuntimeError, match="Cannot create a Window before creating an App"
    ):
        toga.Window()


def test_set_app(window, app):
    """A window's app cannot be reassigned."""
    assert window.app == app

    app2 = toga.App("Test App 2", "org.beeware.toga.test-app-2")
    with pytest.raises(ValueError, match=r"Window is already associated with an App"):
        window.app = app2


def test_set_app_with_content(window, app):
    """If a window has content, the content is assigned to the app."""
    assert window.app == app

    content = toga.Box()
    assert content.app is None

    window.content = content
    assert content.app == app


@pytest.mark.parametrize(
    "value, expected",
    [
        ("New Text", "New Text"),
        ("", "Toga"),
        (None, "Toga"),
        (12345, "12345"),
        ("Contains\nnewline", "Contains"),
    ],
)
def test_title(window, value, expected):
    """The title of the window can be changed."""
    window.title = value
    assert window.title == expected


def test_toolbar_implicit_add(window, app):
    """Adding an item to a toolbar implicitly adds it to the app."""
    cmd1 = toga.Command(None, "Command 1")
    cmd2 = toga.Command(None, "Command 2")

    toolbar = window.toolbar
    assert list(toolbar) == []
    assert list(app.commands) == []

    # Adding a command to the toolbar automatically adds it to the app
    toolbar.add(cmd1)
    assert list(toolbar) == [cmd1]
    assert list(app.commands) == [cmd1]

    # But not vice versa
    app.commands.add(cmd2)
    assert list(toolbar) == [cmd1]
    assert list(app.commands) == [cmd1, cmd2]

    # Adding a command to both places does not cause a duplicate
    app.commands.add(cmd1)
    assert list(toolbar) == [cmd1]
    assert list(app.commands) == [cmd1, cmd2]


def test_change_content(window, app):
    """The content of a window can be changed."""
    assert window.content is None
    assert window.app == app

    # Set the content of the window
    content1 = toga.Box()
    window.content = content1

    # The content has been assigned and refreshed
    assert content1.app == app
    assert content1.window == window
    assert_action_performed_with(window, "set content", widget=content1._impl)
    assert_action_performed(content1, "refresh")

    # Set the content of the window to something new
    content2 = toga.Box()
    window.content = content2

    # The content has been assigned and refreshed
    assert content2.app == app
    assert content2.window == window
    assert_action_performed_with(window, "set content", widget=content2._impl)
    assert_action_performed(content2, "refresh")

    # The original content has been removed
    assert content1.window is None


def test_set_position(window):
    """The position of the window can be set."""
    window.position = (123, 456)

    assert window.position == (123, 456)


def test_set_size(window):
    """The size of the window can be set."""
    window.size = (123, 456)

    assert window.size == (123, 456)


def test_set_size_with_content(window):
    """The size of the window can be set."""
    content = toga.Box()
    window.content = content

    window.size = (123, 456)

    assert window.size == (123, 456)
    assert_action_performed(content, "refresh")


def test_show_hide(window, app):
    """The window can be shown and hidden."""
    assert window.app == app
    window.show()

    # The window has been assigned to the app, and is visible
    assert window.app == app
    assert window in app.windows
    assert_action_performed(window, "show")
    assert window.visible

    # Hide with an explicit call
    window.hide()

    # Window is still assigned to the app, but is not visible
    assert window.app == app
    assert window in app.windows
    assert_action_performed(window, "hide")
    assert not window.visible


def test_hide_show(window, app):
    """The window can be hidden then shown."""
    assert window.app == app
    window.hide()

    # The window has been assigned to the app, and is not visible
    assert window.app == app
    assert window in app.windows
    assert_action_performed(window, "hide")
    assert not window.visible

    # Show with an explicit call
    window.show()

    # Window is still assigned to the app, but is not visible
    assert window.app == app
    assert window in app.windows
    assert_action_performed(window, "show")
    assert window.visible


def test_visibility(window, app):
    """The window can be shown and hidden using the visible property."""
    assert window.app == app
    window.visible = True

    # The window has been assigned to the app, and is visible
    assert window.app == app
    assert window in app.windows
    assert_action_performed(window, "show")
    assert window.visible

    # Hide with an explicit call
    window.visible = False

    # Window is still assigned to the app, but is not visible
    assert window.app == app
    assert window in app.windows
    assert_action_performed(window, "hide")
    assert not window.visible


def test_full_screen(window, app):
    """A window can be set full screen."""
    assert not window.full_screen

    window.full_screen = True
    assert window.full_screen
    assert_action_performed_with(window, "set full screen", full_screen=True)

    window.full_screen = False
    assert not window.full_screen
    assert_action_performed_with(window, "set full screen", full_screen=False)


def test_close_direct(window, app):
    """A window can be closed directly."""
    on_close_handler = Mock(return_value=True)
    window.on_close = on_close_handler

    window.show()
    assert window.app == app
    assert window in app.windows

    # Close the window directly
    window.close()

    # Window has been closed, but the close handler has *not* been invoked.
    assert window.closed
    assert window.app == app
    assert window not in app.windows
    assert_action_performed(window, "close")
    on_close_handler.assert_not_called()


def test_close_no_handler(window, app):
    """A window without a close handler can be closed."""
    window.show()
    assert window.app == app
    assert window in app.windows

    # Close the window
    window._impl.simulate_close()

    # Window has been closed, and is no longer in the app's list of windows.
    assert window.closed
    assert window.app == app
    assert window not in app.windows
    assert_action_performed(window, "close")


def test_close_successful_handler(window, app):
    """A window with a successful close handler can be closed."""
    on_close_handler = Mock(return_value=True)
    window.on_close = on_close_handler

    window.show()
    assert window.app == app
    assert window in app.windows

    # Close the window
    window._impl.simulate_close()

    # Window has been closed, and is no longer in the app's list of windows.
    assert window.closed
    assert window.app == app
    assert window not in app.windows
    assert_action_performed(window, "close")
    on_close_handler.assert_called_once_with(window)


def test_close_rejected_handler(window, app):
    """A window can have a close handler that rejects closing."""
    on_close_handler = Mock(return_value=False)
    window.on_close = on_close_handler

    window.show()
    assert window.app == app
    assert window in app.windows

    # Close the window
    window._impl.simulate_close()

    # Window has *not* been closed
    assert not window.closed
    assert window.app == app
    assert window in app.windows
    assert_action_not_performed(window, "close")
    on_close_handler.assert_called_once_with(window)


def test_as_image(window):
    """A window can be captured as an image."""
    image = window.as_image()
    assert_action_performed(window, "get image data")
    # Don't need to check the raw data; just check it's the right size.
    assert image.size == (318, 346)


def test_screen(window, app):
    """A window can be moved to a different screen."""
    # Cannot actually change window.screen, so just check
    # the window positions as a substitute for moving the
    # window between the screens.
    # `window.screen` will return `Secondary Screen`
    assert window.screen == app.screens[1]
    assert window.position == (100, 100)
    window.screen = app.screens[0]
    assert window.position == (1466, 868)


def test_screen_position(window, app):
    """The window can be relocated using absolute and relative screen positions."""
    # Details about screen layout are in toga_dummy=>app.py=>get_screens()
    initial_position = window.position
    window.position = (-100, -100)
    assert window.position != initial_position
    assert window.position == (-100, -100)
    assert window.screen_position == (1266, 668)

    # Move the window to a new position.
    window.screen_position = (100, 100)
    assert window.position == (-1266, -668)
    assert window.screen_position == (100, 100)


def test_widget_id_reusablity(window, app):
    """Widget IDs can be reused after the associated widget's window is closed."""
    # Common IDs
    CONTENT_WIDGET_ID = "sample_window_content"
    LABEL_WIDGET_ID = "sample_label"

    label_widget = toga.Label(text="Sample Label", id=LABEL_WIDGET_ID)
    second_window_content = toga.Box(id=CONTENT_WIDGET_ID, children=[label_widget])

    third_window_content = toga.Box(children=[])

    # A widget ID is only "used" when it is part of a visible layout;
    # creating a widget and *not* putting it in a layout isn't an problem.
    try:
        new_label_widget = toga.Label(text="New Label", id=LABEL_WIDGET_ID)
    except KeyError:
        pytest.fail("Widget IDs that aren't part of a layout can be re-used.")

    # Create 2 new visible windows
    second_window = toga.Window()
    second_window.show()

    third_window = toga.Window()
    third_window.show()

    # Assign the window content widget to the second window
    second_window.content = second_window_content
    assert CONTENT_WIDGET_ID in app.widgets
    assert LABEL_WIDGET_ID in app.widgets

    # CONTENT_WIDGET_ID is in use, so a widget with that ID can't be assigned to a window.
    with pytest.raises(
        KeyError,
        match=r"There is already a widget with the id 'sample_label'",
    ):
        third_window.content = new_label_widget
    assert CONTENT_WIDGET_ID not in third_window.widgets
    assert LABEL_WIDGET_ID not in third_window.widgets

    # Adding content that has a child with a re-used ID should raise an error
    with pytest.raises(
        KeyError,
        match=r"There is already a widget with the id 'sample_label'",
    ):
        third_window.content = toga.Box(children=[new_label_widget])
    assert CONTENT_WIDGET_ID not in third_window.widgets
    assert LABEL_WIDGET_ID not in third_window.widgets

    # Adding a child with a re-used ID should raise an error.
    third_window.content = third_window_content
    with pytest.raises(
        KeyError,
        match=r"There is already a widget with the id 'sample_label'",
    ):
        third_window_content.add(new_label_widget)
    assert CONTENT_WIDGET_ID not in third_window.widgets
    assert LABEL_WIDGET_ID not in third_window.widgets

    # Creating a new widget with same widget ID should not raise KeyError
    try:
        another_label_widget = toga.Label(text="Another Label", id=LABEL_WIDGET_ID)
    except KeyError:
        pytest.fail("Widget IDs that aren't part of a layout can be re-used.")

    # If a widget using an ID is being *replaced*, the ID can be re-used.
    try:
        second_window.content = another_label_widget
    except KeyError:
        pytest.fail("Widget IDs that are replaced can be re-used.")

    # Close Window 2
    second_window.close()
    assert CONTENT_WIDGET_ID not in app.widgets
    assert LABEL_WIDGET_ID not in app.widgets

    # Now that second_window has been closed, the duplicate ID can be used
    try:
        third_window_content.add(new_label_widget)
    except KeyError:
        pytest.fail("Widget IDs that are replaced can be re-used.")

    third_window.close()


def test_info_dialog(window, app):
    """An info dialog can be shown."""
    on_result_handler = Mock()

    with pytest.warns(
        DeprecationWarning,
        match=r"Synchronous `on_result` handlers have been deprecated;",
    ):
        dialog = window.info_dialog("Title", "Body", on_result=on_result_handler)

    assert dialog.window == window
    assert dialog.app == app

    with pytest.raises(
        RuntimeError,
        match=r"Can't check dialog result directly; use await or an on_result handler",
    ):
        # Perform a synchronous comparison; this will raise a runtime error
        dialog == 1

    async def run_dialog(dialog):
        dialog._impl.simulate_result(None)
        return await dialog

    assert app._impl.loop.run_until_complete(run_dialog(dialog)) is None

    assert_action_performed_with(
        window,
        "show info dialog",
        title="Title",
        message="Body",
    )
    on_result_handler.assert_called_once_with(window, None)


def test_question_dialog(window, app):
    """A question dialog can be shown."""
    on_result_handler = Mock()

    with pytest.warns(
        DeprecationWarning,
        match=r"Synchronous `on_result` handlers have been deprecated;",
    ):
        dialog = window.question_dialog("Title", "Body", on_result=on_result_handler)

    assert dialog.window == window
    assert dialog.app == app

    with pytest.raises(
        RuntimeError,
        match=r"Can't check dialog result directly; use await or an on_result handler",
    ):
        # Perform a synchronous comparison; this will raise a runtime error
        dialog == 1

    async def run_dialog(dialog):
        dialog._impl.simulate_result(True)
        return await dialog

    assert app._impl.loop.run_until_complete(run_dialog(dialog))

    assert_action_performed_with(
        window,
        "show question dialog",
        title="Title",
        message="Body",
    )
    on_result_handler.assert_called_once_with(window, True)


def test_confirm_dialog(window, app):
    """A confirm dialog can be shown."""
    on_result_handler = Mock()

    with pytest.warns(
        DeprecationWarning,
        match=r"Synchronous `on_result` handlers have been deprecated;",
    ):
        dialog = window.confirm_dialog("Title", "Body", on_result=on_result_handler)

    assert dialog.window == window
    assert dialog.app == app

    with pytest.raises(
        RuntimeError,
        match=r"Can't check dialog result directly; use await or an on_result handler",
    ):
        # Perform a synchronous comparison; this will raise a runtime error
        dialog == 1

    async def run_dialog(dialog):
        dialog._impl.simulate_result(True)
        return await dialog

    assert app._impl.loop.run_until_complete(run_dialog(dialog))

    assert_action_performed_with(
        window,
        "show confirm dialog",
        title="Title",
        message="Body",
    )
    on_result_handler.assert_called_once_with(window, True)


def test_error_dialog(window, app):
    """An error dialog can be shown."""
    on_result_handler = Mock()

    with pytest.warns(
        DeprecationWarning,
        match=r"Synchronous `on_result` handlers have been deprecated;",
    ):
        dialog = window.error_dialog("Title", "Body", on_result=on_result_handler)

    assert dialog.window == window
    assert dialog.app == app

    with pytest.raises(
        RuntimeError,
        match=r"Can't check dialog result directly; use await or an on_result handler",
    ):
        # Perform a synchronous comparison; this will raise a runtime error
        dialog == 1

    async def run_dialog(dialog):
        dialog._impl.simulate_result(None)
        return await dialog

    assert app._impl.loop.run_until_complete(run_dialog(dialog)) is None

    assert_action_performed_with(
        window,
        "show error dialog",
        title="Title",
        message="Body",
    )
    on_result_handler.assert_called_once_with(window, None)


def test_stack_trace_dialog(window, app):
    """A stack trace dialog can be shown."""
    on_result_handler = Mock()

    with pytest.warns(
        DeprecationWarning,
        match=r"Synchronous `on_result` handlers have been deprecated;",
    ):
        dialog = window.stack_trace_dialog(
            "Title",
            "Body",
            "The error",
            on_result=on_result_handler,
        )

    assert dialog.window == window
    assert dialog.app == app

    with pytest.raises(
        RuntimeError,
        match=r"Can't check dialog result directly; use await or an on_result handler",
    ):
        # Perform a synchronous comparison; this will raise a runtime error
        dialog == 1

    async def run_dialog(dialog):
        dialog._impl.simulate_result(None)
        return await dialog

    assert app._impl.loop.run_until_complete(run_dialog(dialog)) is None

    assert_action_performed_with(
        window,
        "show stack trace dialog",
        title="Title",
        message="Body",
        content="The error",
        retry=False,
    )
    on_result_handler.assert_called_once_with(window, None)


def test_save_file_dialog(window, app):
    """A save file dialog can be shown."""
    on_result_handler = Mock()

    with pytest.warns(
        DeprecationWarning,
        match=r"Synchronous `on_result` handlers have been deprecated;",
    ):
        dialog = window.save_file_dialog(
            "Title",
            Path("/path/to/initial_file.txt"),
            on_result=on_result_handler,
        )

    assert dialog.window == window
    assert dialog.app == app

    with pytest.raises(
        RuntimeError,
        match=r"Can't check dialog result directly; use await or an on_result handler",
    ):
        # Perform a synchronous comparison; this will raise a runtime error
        dialog == 1

    saved_file = Path("/saved/path/filename.txt")

    async def run_dialog(dialog):
        dialog._impl.simulate_result(saved_file)
        return await dialog

    assert app._impl.loop.run_until_complete(run_dialog(dialog)) is saved_file

    assert_action_performed_with(
        window,
        "show save file dialog",
        title="Title",
        filename="initial_file.txt",
        initial_directory=Path("/path/to"),
        file_types=None,
    )
    on_result_handler.assert_called_once_with(window, saved_file)


def test_save_file_dialog_default_directory(window, app):
    """If no path is provided, a save file dialog will use the default directory."""
    on_result_handler = Mock()

    with pytest.warns(
        DeprecationWarning,
        match=r"Synchronous `on_result` handlers have been deprecated;",
    ):
        dialog = window.save_file_dialog(
            "Title",
            "initial_file.txt",
            file_types=[".txt", ".pdf"],
            on_result=on_result_handler,
        )

    assert dialog.window == window
    assert dialog.app == app

    with pytest.raises(
        RuntimeError,
        match=r"Can't check dialog result directly; use await or an on_result handler",
    ):
        # Perform a synchronous comparison; this will raise a runtime error
        dialog == 1

    saved_file = Path("/saved/path/filename.txt")

    async def run_dialog(dialog):
        dialog._impl.simulate_result(saved_file)
        return await dialog

    assert app._impl.loop.run_until_complete(run_dialog(dialog)) is saved_file

    assert_action_performed_with(
        window,
        "show save file dialog",
        title="Title",
        filename="initial_file.txt",
        initial_directory=None,
        file_types=[".txt", ".pdf"],
    )
    on_result_handler.assert_called_once_with(window, saved_file)


def test_open_file_dialog(window, app):
    """A open file dialog can be shown."""
    on_result_handler = Mock()

    with pytest.warns(
        DeprecationWarning,
        match=r"Synchronous `on_result` handlers have been deprecated;",
    ):
        dialog = window.open_file_dialog(
            "Title",
            "/path/to/folder",
            on_result=on_result_handler,
        )

    assert dialog.window == window
    assert dialog.app == app

    with pytest.raises(
        RuntimeError,
        match=r"Can't check dialog result directly; use await or an on_result handler",
    ):
        # Perform a synchronous comparison; this will raise a runtime error
        dialog == 1

    opened_file = Path("/opened/path/filename.txt")

    async def run_dialog(dialog):
        dialog._impl.simulate_result(opened_file)
        return await dialog

    assert app._impl.loop.run_until_complete(run_dialog(dialog)) is opened_file

    assert_action_performed_with(
        window,
        "show open file dialog",
        title="Title",
        initial_directory=Path("/path/to/folder"),
        file_types=None,
        multiple_select=False,
    )
    on_result_handler.assert_called_once_with(window, opened_file)


def test_open_file_dialog_default_directory(window, app):
    """If no path is provided, a open file dialog will use the default directory."""
    on_result_handler = Mock()

    with pytest.warns(
        DeprecationWarning,
        match=r"Synchronous `on_result` handlers have been deprecated;",
    ):
        dialog = window.open_file_dialog(
            "Title",
            file_types=[".txt", ".pdf"],
            multiple_select=True,
            on_result=on_result_handler,
        )

    assert dialog.window == window
    assert dialog.app == app

    with pytest.raises(
        RuntimeError,
        match=r"Can't check dialog result directly; use await or an on_result handler",
    ):
        # Perform a synchronous comparison; this will raise a runtime error
        dialog == 1

    opened_files = [
        Path("/opened/path/filename.txt"),
        Path("/other/path/filename2.txt"),
    ]

    async def run_dialog(dialog):
        dialog._impl.simulate_result(opened_files)
        return await dialog

    assert app._impl.loop.run_until_complete(run_dialog(dialog)) is opened_files

    assert_action_performed_with(
        window,
        "show open file dialog",
        title="Title",
        initial_directory=None,
        file_types=[".txt", ".pdf"],
        multiple_select=True,
    )
    on_result_handler.assert_called_once_with(window, opened_files)


def test_select_folder_dialog(window, app):
    """A select folder dialog can be shown."""
    on_result_handler = Mock()

    with pytest.warns(
        DeprecationWarning,
        match=r"Synchronous `on_result` handlers have been deprecated;",
    ):
        dialog = window.select_folder_dialog(
            "Title",
            Path("/path/to/folder"),
            on_result=on_result_handler,
        )

    assert dialog.window == window
    assert dialog.app == app

    with pytest.raises(
        RuntimeError,
        match=r"Can't check dialog result directly; use await or an on_result handler",
    ):
        # Perform a synchronous comparison; this will raise a runtime error
        dialog == 1

    opened_file = Path("/opened/path/filename.txt")

    async def run_dialog(dialog):
        dialog._impl.simulate_result(opened_file)
        return await dialog

    assert app._impl.loop.run_until_complete(run_dialog(dialog)) is opened_file

    assert_action_performed_with(
        window,
        "show select folder dialog",
        title="Title",
        initial_directory=Path("/path/to/folder"),
        multiple_select=False,
    )
    on_result_handler.assert_called_once_with(window, opened_file)


def test_select_folder_dialog_default_directory(window, app):
    """If no path is provided, a select folder dialog will use the default directory."""
    on_result_handler = Mock()

    with pytest.warns(
        DeprecationWarning,
        match=r"Synchronous `on_result` handlers have been deprecated;",
    ):
        dialog = window.select_folder_dialog(
            "Title",
            multiple_select=True,
            on_result=on_result_handler,
        )

    assert dialog.window == window
    assert dialog.app == app

    with pytest.raises(
        RuntimeError,
        match=r"Can't check dialog result directly; use await or an on_result handler",
    ):
        # Perform a synchronous comparison; this will raise a runtime error
        dialog == 1

    opened_files = [
        Path("/opened/path/filename.txt"),
        Path("/other/path/filename2.txt"),
    ]

    async def run_dialog(dialog):
        dialog._impl.simulate_result(opened_files)
        return await dialog

    assert app._impl.loop.run_until_complete(run_dialog(dialog)) is opened_files

    assert_action_performed_with(
        window,
        "show select folder dialog",
        title="Title",
        initial_directory=None,
        multiple_select=True,
    )
    on_result_handler.assert_called_once_with(window, opened_files)


def test_deprecated_names_open_file_dialog(window, app):
    """Deprecated names still work on open file dialogs."""
    on_result_handler = Mock()

    with pytest.warns(
        DeprecationWarning,
        match=r"open_file_dialog\(multiselect\) has been renamed multiple_select",
    ), pytest.warns(
        DeprecationWarning,
        match=r"Synchronous `on_result` handlers have been deprecated; use `await` on the asynchronous result",
    ):
        dialog = window.open_file_dialog(
            "Title",
            "/path/to/folder",
            multiselect=True,
            on_result=on_result_handler,
        )

    opened_files = [Path("/opened/path/filename.txt")]

    async def run_dialog(dialog):
        dialog._impl.simulate_result(opened_files)
        return await dialog

    assert app._impl.loop.run_until_complete(run_dialog(dialog)) is opened_files

    assert_action_performed_with(
        window,
        "show open file dialog",
        title="Title",
        initial_directory=Path("/path/to/folder"),
        file_types=None,
        multiple_select=True,
    )
    on_result_handler.assert_called_once_with(window, opened_files)


def test_deprecated_names_select_folder_dialog(window, app):
    """Deprecated names still work on open file dialogs."""
    on_result_handler = Mock()

    with pytest.warns(
        DeprecationWarning,
        match=r"select_folder_dialog\(multiselect\) has been renamed multiple_select",
    ), pytest.warns(
        DeprecationWarning,
        match=r"Synchronous `on_result` handlers have been deprecated; use `await` on the asynchronous result",
    ):
        dialog = window.select_folder_dialog(
            "Title",
            "/path/to/folder",
            multiselect=True,
            on_result=on_result_handler,
        )

    opened_files = [Path("/opened/path")]

    async def run_dialog(dialog):
        dialog._impl.simulate_result(opened_files)
        return await dialog

    assert app._impl.loop.run_until_complete(run_dialog(dialog)) is opened_files

    assert_action_performed_with(
        window,
        "show select folder dialog",
        title="Title",
        initial_directory=Path("/path/to/folder"),
        multiple_select=True,
    )
    on_result_handler.assert_called_once_with(window, opened_files)


def test_deprecated_names_resizeable():
    """Deprecated spelling of resizable still works."""
    with pytest.warns(
        DeprecationWarning,
        match=r"Window.resizeable has been renamed Window.resizable",
    ):
        window = toga.Window(title="Deprecated", resizeable=True)

    with pytest.warns(
        DeprecationWarning,
        match=r"Window.resizeable has been renamed Window.resizable",
    ):
        assert window.resizeable


def test_deprecated_names_closeable():
    """Deprecated spelling of closable still works."""
    with pytest.warns(
        DeprecationWarning,
        match=r"Window.closeable has been renamed Window.closable",
    ):
        window = toga.Window(title="Deprecated", closeable=True)

    with pytest.warns(
        DeprecationWarning,
        match=r"Window.closeable has been renamed Window.closable",
    ):
        assert window.closeable



from __future__ import annotations

import warnings
from builtins import id as identifier
from collections.abc import Mapping, MutableSet
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    ItemsView,
    Iterator,
    KeysView,
    Literal,
    Protocol,
    TypeVar,
    ValuesView,
    overload,
)

from toga.command import Command, CommandSet
from toga.handlers import AsyncResult, wrapped_handler
from toga.images import Image
from toga.platform import get_platform_factory

if TYPE_CHECKING:
    from toga.app import App
    from toga.images import ImageT
    from toga.screens import Screen
    from toga.widgets.base import Widget


class FilteredWidgetRegistry:
    # A class that exposes a mapping lookup interface, filtered to widgets from a single
    # window. The underlying data store is on the app.

    def __init__(self, window):
        self._window = window

    def __len__(self) -> int:
        return len(list(self.items()))

    def __getitem__(self, key: str) -> Widget:
        item = self._window.app.widgets[key]
        if item.window != self._window:
            raise KeyError(key)
        return item

    def __contains__(self, key: str) -> bool:
        try:
            item = self._window.app.widgets[key]
            return item.window == self._window
        except KeyError:
            return False

    def __iter__(self) -> Iterator[Widget]:
        return iter(self.values())

    def __repr__(self) -> str:
        return "{" + ", ".join(f"{k!r}: {v!r}" for k, v in sorted(self.items())) + "}"

    def items(self) -> ItemsView:
        for item in self._window.app.widgets.items():
            if item[1].window == self._window:
                yield item

    def keys(self) -> KeysView:
        for item in self._window.app.widgets.items():
            if item[1].window == self._window:
                yield item[0]

    def values(self) -> ValuesView:
        for item in self._window.app.widgets.items():
            if item[1].window == self._window:
                yield item[1]


class OnCloseHandler(Protocol):
    def __call__(self, window: Window, **kwargs: Any) -> bool:
        """A handler to invoke when a window is about to close.

        The return value of this callback controls whether the window is allowed to close.
        This can be used to prevent a window closing with unsaved changes, etc.

        :param window: The window instance that is closing.
        :param kwargs: Ensures compatibility with additional arguments introduced in
            future versions.
        :returns: ``True`` if the window is allowed to close; ``False`` if the window is not
            allowed to close.
        """
        ...


T = TypeVar("T")


class DialogResultHandler(Protocol[T]):
    def __call__(self, window: Window, result: T, **kwargs: Any) -> None:
        """A handler to invoke when a dialog is closed.

        :param window: The window that opened the dialog.
        :param result: The result returned by the dialog.
        :param kwargs: Ensures compatibility with additional arguments introduced in
            future versions.
        """
        ...


class Dialog(AsyncResult):
    RESULT_TYPE = "dialog"

    def __init__(self, window: Window, on_result: DialogResultHandler[Any]):
        super().__init__(on_result=on_result)
        self.window = window
        self.app = window.app


class Window:
    _WINDOW_CLASS = "Window"

    def __init__(
        self,
        id: str | None = None,
        title: str | None = None,
        position: tuple[int, int] = (100, 100),
        size: tuple[int, int] = (640, 480),
        resizable: bool = True,
        closable: bool = True,
        minimizable: bool = True,
        on_close: OnCloseHandler | None = None,
        resizeable=None,  # DEPRECATED
        closeable=None,  # DEPRECATED
    ) -> None:
        """Create a new Window.

        :param id: A unique identifier for the window. If not provided, one will be
            automatically generated.
        :param title: Title for the window. Defaults to "Toga".
        :param position: Position of the window, as a tuple of ``(x, y)`` coordinates,
            in :ref:`CSS pixels <css-units>`.
        :param size: Size of the window, as a tuple of ``(width, height)``, in :ref:`CSS
            pixels <css-units>`.
        :param resizable: Can the window be resized by the user?
        :param closable: Can the window be closed by the user?
        :param minimizable: Can the window be minimized by the user?
        :param on_close: The initial :any:`on_close` handler.
        :param resizeable: **DEPRECATED** - Use ``resizable``.
        :param closeable: **DEPRECATED** - Use ``closable``.
        """
        ######################################################################
        # 2023-08: Backwards compatibility
        ######################################################################
        if resizeable is not None:
            warnings.warn(
                "Window.resizeable has been renamed Window.resizable",
                DeprecationWarning,
            )
            resizable = resizeable

        if closeable is not None:
            warnings.warn(
                "Window.closeable has been renamed Window.closable",
                DeprecationWarning,
            )
            closable = closeable
        ######################################################################
        # End backwards compatibility
        ######################################################################

        # Needs to be a late import to avoid circular dependencies.
        from toga import App

        self._id = str(id if id else identifier(self))
        self._impl = None
        self._content = None
        self._is_full_screen = False
        self._closed = False

        self._resizable = resizable
        self._closable = closable
        self._minimizable = minimizable

        self.factory = get_platform_factory()
        self._impl = getattr(self.factory, self._WINDOW_CLASS)(
            interface=self,
            title=title if title else self._default_title,
            position=position,
            size=size,
        )

        # Add the window to the app
        self._app = None
        if App.app is None:
            raise RuntimeError("Cannot create a Window before creating an App")
        App.app.windows.add(self)

        # Create a toolbar that is linked to the app
        self._toolbar = CommandSet(on_change=self._impl.create_toolbar, app=self._app)

        self.on_close = on_close

    def __lt__(self, other) -> bool:
        return self.id < other.id

    ######################################################################
    # Window properties
    ######################################################################

    @property
    def app(self) -> App:
        """The :any:`App` that this window belongs to (read-only).

        New windows are automatically associated with the currently active app."""
        return self._app

    @app.setter
    def app(self, app: App) -> None:
        if self._app:
            raise ValueError("Window is already associated with an App")

        self._app = app
        self._impl.set_app(app._impl)

    @property
    def closable(self) -> bool:
        """Can the window be closed by the user?"""
        return self._closable

    @property
    def id(self) -> str:
        """A unique identifier for the window."""
        return self._id

    @property
    def minimizable(self) -> bool:
        """Can the window be minimized by the user?"""
        return self._minimizable

    @property
    def resizable(self) -> bool:
        """Can the window be resized by the user?"""
        return self._resizable

    @property
    def _default_title(self) -> str:
        return "Toga"

    @property
    def title(self) -> str:
        """Title of the window. If no title is provided, the title will default to
        "Toga"."""
        return self._impl.get_title()

    @title.setter
    def title(self, title: str) -> None:
        if not title:
            title = self._default_title

        self._impl.set_title(str(title).split("\n")[0])

    ######################################################################
    # Window lifecycle
    ######################################################################

    def close(self) -> None:
        """Close the window.

        This *does not* invoke the ``on_close`` handler; the window will be immediately
        and unconditionally closed.

        Once a window has been closed, it *cannot* be reused. The behavior of any method
        or property on a :class:`~toga.Window` instance after it has been closed is
        undefined, except for :attr:`closed` which can be used to check if the window
        was closed.
        """
        if self.content:
            self.content.window = None
        self.app.windows.discard(self)
        self._impl.close()
        self._closed = True

    @property
    def closed(self) -> bool:
        """Whether the window was closed."""
        return self._closed

    def show(self) -> None:
        """Show the window. If the window is already visible, this method has no
        effect."""
        self._impl.show()

    ######################################################################
    # Window content and resources
    ######################################################################

    @property
    def content(self) -> Widget | None:
        """Content of the window. On setting, the content is added to the same app as
        the window."""
        return self._content

    @content.setter
    def content(self, widget: Widget) -> None:
        # Set window of old content to None
        if self._content:
            self._content.window = None

        # Assign the content widget to the same app as the window.
        widget.app = self.app

        # Assign the content widget to the window.
        widget.window = self

        # Track our new content
        self._content = widget

        # Manifest the widget
        self._impl.set_content(widget._impl)

        # Update the geometry of the widget
        widget.refresh()

    @property
    def toolbar(self) -> MutableSet[Command]:
        """Toolbar for the window."""
        return self._toolbar

    @property
    def widgets(self) -> Mapping[str, Widget]:
        """The widgets contained in the window.

        Can be used to look up widgets by ID (e.g., ``window.widgets["my_id"]``).
        """
        return FilteredWidgetRegistry(self)

    ######################################################################
    # Window size
    ######################################################################

    @property
    def size(self) -> tuple[int, int]:
        """Size of the window, as a tuple of ``(width, height)``, in
        :ref:`CSS pixels <css-units>`."""
        return self._impl.get_size()

    @size.setter
    def size(self, size: tuple[int, int]) -> None:
        self._impl.set_size(size)
        if self.content:
            self.content.refresh()

    ######################################################################
    # Window position
    ######################################################################

    @property
    def position(self) -> tuple[int, int]:
        """Absolute position of the window, as a ``(x, y)`` tuple coordinates, in
        :ref:`CSS pixels <css-units>`.

        The origin is the top left corner of the primary screen.
        """
        absolute_origin = self._app.screens[0].origin
        absolute_window_position = self._impl.get_position()

        window_position = (
            absolute_window_position[0] - absolute_origin[0],
            absolute_window_position[1] - absolute_origin[1],
        )
        return window_position

    @position.setter
    def position(self, position: tuple[int, int]) -> None:
        absolute_origin = self._app.screens[0].origin
        absolute_new_position = (
            position[0] + absolute_origin[0],
            position[1] + absolute_origin[1],
        )
        self._impl.set_position(absolute_new_position)

    @property
    def screen(self) -> Screen:
        """Instance of the :class:`toga.Screen` on which this window is present."""
        return self._impl.get_current_screen().interface

    @screen.setter
    def screen(self, app_screen: Screen) -> None:
        original_window_location = self.position
        original_origin = self.screen.origin
        new_origin = app_screen.origin
        x = original_window_location[0] - original_origin[0] + new_origin[0]
        y = original_window_location[1] - original_origin[1] + new_origin[1]

        self._impl.set_position((x, y))

    @property
    def screen_position(self) -> tuple[int, int]:
        """Position of the window with respect to current screen, as a ``(x, y)`` tuple."""
        current_relative_position = (
            self.position[0] - self.screen.origin[0],
            self.position[1] - self.screen.origin[1],
        )
        return current_relative_position

    @screen_position.setter
    def screen_position(self, position: tuple[int, int]) -> None:
        new_relative_position = (
            position[0] + self.screen.origin[0],
            position[1] + self.screen.origin[1],
        )
        self._impl.set_position(new_relative_position)

    ######################################################################
    # Window visibility
    ######################################################################

    def hide(self) -> None:
        """Hide the window. If the window is already hidden, this method has no
        effect."""
        self._impl.hide()

    @property
    def visible(self) -> bool:
        """Is the window visible?"""
        return self._impl.get_visible()

    @visible.setter
    def visible(self, visible: bool) -> None:
        if visible:
            self.show()
        else:
            self.hide()

    ######################################################################
    # Window state
    ######################################################################

    @property
    def full_screen(self) -> bool:
        """Is the window in full screen mode?

        Full screen mode is *not* the same as "maximized". A full screen window
        has no title bar, toolbar or window controls; some or all of these
        items may be visible on a maximized window. A good example of "full screen"
        mode is a slideshow app in presentation mode - the only visible content is
        the slide.
        """
        return self._is_full_screen

    @full_screen.setter
    def full_screen(self, is_full_screen: bool) -> None:
        self._is_full_screen = is_full_screen
        self._impl.set_full_screen(is_full_screen)

    ######################################################################
    # Window capabilities
    ######################################################################

    def as_image(self, format: type[ImageT] = Image) -> ImageT:
        """Render the current contents of the window as an image.

        :param format: Format to provide. Defaults to :class:`~toga.images.Image`; also
            supports :any:`PIL.Image.Image` if Pillow is installed, as well as any image
            types defined by installed :doc:`image format plugins
            </reference/plugins/image_formats>`.
        :returns: An image containing the window content, in the format requested.
        """
        return Image(self._impl.get_image_data()).as_format(format)

    ######################################################################
    # Window events
    ######################################################################

    @property
    def on_close(self) -> OnCloseHandler:
        """The handler to invoke if the user attempts to close the window."""
        return self._on_close

    @on_close.setter
    def on_close(self, handler: OnCloseHandler | None) -> None:
        def cleanup(window: Window, should_close: bool) -> None:
            if should_close or handler is None:
                window.close()

        self._on_close = wrapped_handler(self, handler, cleanup=cleanup)

    ######################################################################
    # Dialogs
    ######################################################################

    def info_dialog(
        self,
        title: str,
        message: str,
        on_result: DialogResultHandler[None] | None = None,
    ) -> Dialog:
        """Ask the user to acknowledge some information.

        Presents as a dialog with a single "OK" button to close the dialog.

        **This is an asynchronous method**. If you invoke this method in synchronous
        context, it will show the dialog, but will return *immediately*. The object
        returned by this method can be awaited to obtain the result of the dialog.

        :param title: The title of the dialog window.
        :param message: The message to display.
        :param on_result: **DEPRECATED** ``await`` the return value of this method.
        :returns: An awaitable Dialog object. The Dialog object returns ``None`` when
            the user presses the 'OK' button.
        """
        dialog = Dialog(
            self,
            on_result=wrapped_handler(self, on_result) if on_result else None,
        )
        self.factory.dialogs.InfoDialog(dialog, title, message)
        return dialog

    def question_dialog(
        self,
        title: str,
        message: str,
        on_result: DialogResultHandler[bool] | None = None,
    ) -> Dialog:
        """Ask the user a yes/no question.

        Presents as a dialog with "Yes" and "No" buttons.

        **This is an asynchronous method**. If you invoke this method in synchronous
        context, it will show the dialog, but will return *immediately*. The object
        returned by this method can be awaited to obtain the result of the dialog.

        :param title: The title of the dialog window.
        :param message: The question to be answered.
        :param on_result: **DEPRECATED** ``await`` the return value of this method.
        :returns: An awaitable Dialog object. The Dialog object returns ``True`` when
            the "Yes" button is pressed, ``False`` when the "No" button is pressed.
        """
        dialog = Dialog(
            self,
            on_result=wrapped_handler(self, on_result) if on_result else None,
        )
        self.factory.dialogs.QuestionDialog(dialog, title, message)
        return dialog

    def confirm_dialog(
        self,
        title: str,
        message: str,
        on_result: DialogResultHandler[bool] | None = None,
    ) -> Dialog:
        """Ask the user to confirm if they wish to proceed with an action.

        Presents as a dialog with "Cancel" and "OK" buttons (or whatever labels are
        appropriate on the current platform).

        **This is an asynchronous method**. If you invoke this method in synchronous
        context, it will show the dialog, but will return *immediately*. The object
        returned by this method can be awaited to obtain the result of the dialog.

        :param title: The title of the dialog window.
        :param message: A message describing the action to be confirmed.
        :param on_result: **DEPRECATED** ``await`` the return value of this method.
        :returns: An awaitable Dialog object. The Dialog object returns ``True`` when
            the "OK" button is pressed, ``False`` when the "Cancel" button is pressed.
        """
        dialog = Dialog(
            self,
            on_result=wrapped_handler(self, on_result) if on_result else None,
        )
        self.factory.dialogs.ConfirmDialog(dialog, title, message)
        return dialog

    def error_dialog(
        self,
        title: str,
        message: str,
        on_result: DialogResultHandler[None] | None = None,
    ) -> Dialog:
        """Ask the user to acknowledge an error state.

        Presents as an error dialog with an "OK" button to close the dialog.

        **This is an asynchronous method**. If you invoke this method in synchronous
        context, it will show the dialog, but will return *immediately*. The object
        returned by this method can be awaited to obtain the result of the dialog.

        :param title: The title of the dialog window.
        :param message: The error message to display.
        :param on_result: **DEPRECATED** ``await`` the return value of this method.
        :returns: An awaitable Dialog object. The Dialog object returns ``None`` when
            the user presses the "OK" button.
        """
        dialog = Dialog(
            self,
            on_result=wrapped_handler(self, on_result) if on_result else None,
        )
        self.factory.dialogs.ErrorDialog(dialog, title, message)
        return dialog

    @overload
    def stack_trace_dialog(
        self,
        title: str,
        message: str,
        content: str,
        retry: Literal[False] = False,
        on_result: DialogResultHandler[None] | None = None,
    ) -> Dialog: ...

    @overload
    def stack_trace_dialog(
        self,
        title: str,
        message: str,
        content: str,
        retry: Literal[True] = False,
        on_result: DialogResultHandler[bool] | None = None,
    ) -> Dialog: ...

    @overload
    def stack_trace_dialog(
        self,
        title: str,
        message: str,
        content: str,
        retry: bool = False,
        on_result: DialogResultHandler[bool | None] | None = None,
    ) -> Dialog: ...

    def stack_trace_dialog(
        self,
        title: str,
        message: str,
        content: str,
        retry: bool = False,
        on_result: DialogResultHandler[bool | None] | None = None,
    ) -> Dialog:
        """Open a dialog to display a large block of text, such as a stack trace.

        **This is an asynchronous method**. If you invoke this method in synchronous
        context, it will show the dialog, but will return *immediately*. The object
        returned by this method can be awaited to obtain the result of the dialog.

        :param title: The title of the dialog window.
        :param message: Contextual information about the source of the stack trace.
        :param content: The stack trace, pre-formatted as a multi-line string.
        :param retry: If true, the user will be given options to "Retry" or "Quit"; if
            false, a single option to acknowledge the error will be displayed.
        :param on_result: **DEPRECATED** ``await`` the return value of this method.
        :returns: An awaitable Dialog object. If ``retry`` is true, the Dialog object
            returns ``True`` when the user selects "Retry", and ``False`` when they
            select "Quit". If ``retry`` is false, the Dialog object returns ``None``.
        """
        dialog = Dialog(
            self,
            on_result=wrapped_handler(self, on_result) if on_result else None,
        )
        self.factory.dialogs.StackTraceDialog(
            dialog,
            title,
            message=message,
            content=content,
            retry=retry,
        )
        return dialog

    def save_file_dialog(
        self,
        title: str,
        suggested_filename: Path | str,
        file_types: list[str] | None = None,
        on_result: DialogResultHandler[Path | None] | None = None,
    ) -> Dialog:
        """Prompt the user for a location to save a file.

        This dialog is not currently supported on Android or iOS.

        **This is an asynchronous method**. If you invoke this method in synchronous
        context, it will show the dialog, but will return *immediately*. The object
        returned by this method can be awaited to obtain the result of the dialog.

        :param title: The title of the dialog window
        :param suggested_filename: A default filename
        :param file_types: The allowed filename extensions, without leading dots. If not
            provided, any extension will be allowed.
        :param on_result: **DEPRECATED** ``await`` the return value of this method.
        :returns: An awaitable Dialog object. The Dialog object returns a path object
            for the selected file location, or ``None`` if the user cancelled the save
            operation.
        """
        dialog = Dialog(
            self,
            on_result=wrapped_handler(self, on_result) if on_result else None,
        )
        # Convert suggested filename to a path (if it isn't already),
        # and break it into a filename and a directory
        suggested_path = Path(suggested_filename)
        initial_directory = suggested_path.parent
        if initial_directory == Path("."):
            initial_directory = None
        filename = suggested_path.name

        self.factory.dialogs.SaveFileDialog(
            dialog,
            title,
            filename=filename,
            initial_directory=initial_directory,
            file_types=file_types,
        )
        return dialog

    @overload
    def open_file_dialog(
        self,
        title: str,
        initial_directory: Path | str | None = None,
        file_types: list[str] | None = None,
        multiple_select: Literal[False] = False,
        on_result: DialogResultHandler[Path | None] | None = None,
        multiselect=None,  # DEPRECATED
    ) -> Dialog: ...

    @overload
    def open_file_dialog(
        self,
        title: str,
        initial_directory: Path | str | None = None,
        file_types: list[str] | None = None,
        multiple_select: Literal[True] = True,
        on_result: DialogResultHandler[list[Path] | None] | None = None,
        multiselect=None,  # DEPRECATED
    ) -> Dialog: ...

    @overload
    def open_file_dialog(
        self,
        title: str,
        initial_directory: Path | str | None = None,
        file_types: list[str] | None = None,
        multiple_select: bool = False,
        on_result: DialogResultHandler[list[Path] | Path | None] | None = None,
        multiselect=None,  # DEPRECATED
    ) -> Dialog: ...

    def open_file_dialog(
        self,
        title: str,
        initial_directory: Path | str | None = None,
        file_types: list[str] | None = None,
        multiple_select: bool = False,
        on_result: DialogResultHandler[list[Path] | Path | None] | None = None,
        multiselect=None,  # DEPRECATED
    ) -> Dialog:
        """Prompt the user to select a file (or files) to open.

        This dialog is not currently supported on Android or iOS.

        **This is an asynchronous method**. If you invoke this method in synchronous
        context, it will show the dialog, but will return *immediately*. The object
        returned by this method can be awaited to obtain the result of the dialog.

        :param title: The title of the dialog window
        :param initial_directory: The initial folder in which to open the dialog. If
            ``None``, use the default location provided by the operating system (which
            will often be the last used location)
        :param file_types: The allowed filename extensions, without leading dots. If not
            provided, all files will be shown.
        :param multiple_select: If True, the user will be able to select multiple files;
            if False, the selection will be restricted to a single file.
        :param on_result: **DEPRECATED** ``await`` the return value of this method.
        :param multiselect: **DEPRECATED** Use ``multiple_select``.
        :returns: An awaitable Dialog object. The Dialog object returns a list of
            ``Path`` objects if ``multiple_select`` is ``True``, or a single ``Path``
            otherwise. Returns ``None`` if the open operation is cancelled by the user.
        """
        ######################################################################
        # 2023-08: Backwards compatibility
        ######################################################################
        if multiselect is not None:
            warnings.warn(
                "open_file_dialog(multiselect) has been renamed multiple_select",
                DeprecationWarning,
            )
            multiple_select = multiselect
        ######################################################################
        # End Backwards compatibility
        ######################################################################

        dialog = Dialog(
            self,
            on_result=wrapped_handler(self, on_result) if on_result else None,
        )
        self.factory.dialogs.OpenFileDialog(
            dialog,
            title,
            initial_directory=Path(initial_directory) if initial_directory else None,
            file_types=file_types,
            multiple_select=multiple_select,
        )
        return dialog

    @overload
    def select_folder_dialog(
        self,
        title: str,
        initial_directory: Path | str | None = None,
        multiple_select: Literal[False] = False,
        on_result: DialogResultHandler[Path | None] | None = None,
        multiselect=None,  # DEPRECATED
    ) -> Dialog: ...

    @overload
    def select_folder_dialog(
        self,
        title: str,
        initial_directory: Path | str | None = None,
        multiple_select: Literal[True] = True,
        on_result: DialogResultHandler[list[Path] | None] | None = None,
        multiselect=None,  # DEPRECATED
    ) -> Dialog: ...

    @overload
    def select_folder_dialog(
        self,
        title: str,
        initial_directory: Path | str | None = None,
        multiple_select: bool = False,
        on_result: DialogResultHandler[list[Path] | Path | None] | None = None,
        multiselect=None,  # DEPRECATED
    ) -> Dialog: ...

    def select_folder_dialog(
        self,
        title: str,
        initial_directory: Path | str | None = None,
        multiple_select: bool = False,
        on_result: DialogResultHandler[list[Path] | Path | None] | None = None,
        multiselect=None,  # DEPRECATED
    ) -> Dialog:
        """Prompt the user to select a directory (or directories).

        This dialog is not currently supported on Android or iOS.

        **This is an asynchronous method**. If you invoke this method in synchronous
        context, it will show the dialog, but will return *immediately*. The object
        returned by this method can be awaited to obtain the result of the dialog.

        :param title: The title of the dialog window
        :param initial_directory: The initial folder in which to open the dialog. If
            ``None``, use the default location provided by the operating system (which
            will often be "last used location")
        :param multiple_select: If True, the user will be able to select multiple
            directories; if False, the selection will be restricted to a single
            directory. This option is not supported on WinForms.
        :param on_result: **DEPRECATED** ``await`` the return value of this method.
        :param multiselect: **DEPRECATED** Use ``multiple_select``.
        :returns: An awaitable Dialog object. The Dialog object returns a list of
            ``Path`` objects if ``multiple_select`` is ``True``, or a single ``Path``
            otherwise. Returns ``None`` if the open operation is cancelled by the user.
        """
        ######################################################################
        # 2023-08: Backwards compatibility
        ######################################################################
        if multiselect is not None:
            warnings.warn(
                "select_folder_dialog(multiselect) has been renamed multiple_select",
                DeprecationWarning,
            )
            multiple_select = multiselect
        ######################################################################
        # End Backwards compatibility
        ######################################################################

        dialog = Dialog(
            self,
            on_result=wrapped_handler(self, on_result) if on_result else None,
        )
        self.factory.dialogs.SelectFolderDialog(
            dialog,
            title,
            initial_directory=Path(initial_directory) if initial_directory else None,
            multiple_select=multiple_select,
        )
        return dialog

    ######################################################################
    # 2023-08: Backwards compatibility
    ######################################################################

    @property
    def resizeable(self) -> bool:
        """**DEPRECATED** Use :attr:`resizable`"""
        warnings.warn(
            "Window.resizeable has been renamed Window.resizable",
            DeprecationWarning,
        )
        return self._resizable

    @property
    def closeable(self) -> bool:
        """**DEPRECATED** Use :attr:`closable`"""
        warnings.warn(
            "Window.closeable has been renamed Window.closable",
            DeprecationWarning,
        )
        return self._closable



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Protocol

import toga
from toga.handlers import wrapped_handler

from .base import Widget

if TYPE_CHECKING:
    from toga.icons import IconContent


class OnPressHandler(Protocol):
    def __call__(self, widget: Button, **kwargs: Any) -> None:
        """A handler that will be invoked when a button is pressed.

        :param widget: The button that was pressed.
        :param kwargs: Ensures compatibility with arguments added in future versions.
        """
        ...


class Button(Widget):
    def __init__(
        self,
        text: str | None = None,
        icon: IconContent | None = None,
        id: str | None = None,
        style=None,
        on_press: OnPressHandler | None = None,
        enabled: bool = True,
    ):
        """Create a new button widget.

        :param text: The text to display on the button.
        :param icon: The icon to display on the button. Can be specified as any valid
            :any:`icon content <IconContent>`.
        :param id: The ID for the widget.
        :param style: A style object. If no style is provided, a default style will be
            applied to the widget.
        :param on_press: A handler that will be invoked when the button is pressed.
        :param enabled: Is the button enabled (i.e., can it be pressed?). Optional; by
            default, buttons are created in an enabled state.
        """
        super().__init__(id=id, style=style)

        # Create a platform specific implementation of a Button
        self._impl = self.factory.Button(interface=self)

        # Set a dummy handler before installing the actual on_press, because we do not want
        # on_press triggered by the initial value being set
        self.on_press = None

        # Set the content of the button - either an icon, or text, but not both.
        if icon:
            if text is not None:
                raise ValueError("Cannot specify both text and an icon")
            else:
                self.icon = icon
        else:
            self.text = text

        self.on_press = on_press
        self.enabled = enabled

    @property
    def text(self) -> str:
        """The text displayed on the button.

        ``None``, and the Unicode codepoint U+200B (ZERO WIDTH SPACE), will be
        interpreted and returned as an empty string. Any other object will be converted
        to a string using ``str()``.

        Only one line of text can be displayed. Any content after the first newline will
        be ignored.

        If the button is currently displaying an icon, and text is assigned, the icon
        will be replaced by the new text.

        If the button is currently displaying an icon, the empty string will be
        returned.
        """
        return self._impl.get_text()

    @text.setter
    def text(self, value: str | None) -> None:
        if value is None or value == "\u200B":
            value = ""
        else:
            # Button text can't include line breaks. Strip any content
            # after a line break (if provided)
            value = str(value).split("\n")[0]

        self._impl.set_text(value)
        self._impl.set_icon(None)
        self.refresh()

    @property
    def icon(self) -> toga.Icon | None:
        """The icon displayed on the button.

        Can be specified as any valid :any:`icon content <IconContent>`.

        If the button is currently displaying text, and an icon is assigned, the text
        will be replaced by the new icon.

        If ``None`` is assigned as an icon, the button will become a text button with an
        empty label.

        Returns ``None`` if the button is currently displaying text.
        """
        return self._impl.get_icon()

    @icon.setter
    def icon(self, value: IconContent | None) -> None:
        if isinstance(value, toga.Icon):
            icon = value
            text = ""
        elif value is None:
            if self.icon is None:
                # Already a null icon; nothing changes.
                return
            else:
                icon = None
                text = self._impl.get_text()
        else:
            icon = toga.Icon(value)
            text = ""

        self._impl.set_icon(icon)
        self._impl.set_text(text)
        self.refresh()

    @property
    def on_press(self) -> OnPressHandler:
        """The handler to invoke when the button is pressed."""
        return self._on_press

    @on_press.setter
    def on_press(self, handler):
        self._on_press = wrapped_handler(self, handler)



from __future__ import annotations

from .base import Widget


class Label(Widget):
    def __init__(
        self,
        text: str,
        id=None,
        style=None,
    ):
        """Create a new text label.

        :param text: Text of the label.
        :param id: The ID for the widget.
        :param style: A style object. If no style is provided, a default style
            will be applied to the widget.
        """
        super().__init__(id=id, style=style)

        # Create a platform specific implementation of a Label
        self._impl = self.factory.Label(interface=self)

        self.text = text

    def focus(self):
        "No-op; Label cannot accept input focus"
        pass

    @property
    def text(self) -> str:
        """The text displayed by the label.

        ``None``, and the Unicode codepoint U+200B (ZERO WIDTH SPACE), will be
        interpreted and returned as an empty string. Any other object will be
        converted to a string using ``str()``.
        """
        return self._impl.get_text()

    @text.setter
    def text(self, value):
        if value is None or value == "\u200B":
            text = ""
        else:
            text = str(value)

        self._impl.set_text(text)
        self.refresh()



import pytest

from toga import validators


@pytest.mark.parametrize(
    "value, kwargs, error",
    [
        # Exact minimum size
        ("exact", dict(length=5), None),
        # Exceeds minimum size
        ("this is a long string", dict(length=5), None),
        # Allow empty strings
        ("", dict(length=5), None),
        # Too short
        (
            "bad!",
            dict(length=5),
            "Input is too short (length should be at least 5)",
        ),
        # Too short, different length
        (
            "this is a long string",
            dict(length=25),
            "Input is too short (length should be at least 25)",
        ),
        # Custom error message
        (
            "bad",
            dict(length=5, error_message="Badness"),
            "Badness",
        ),
        # Don't allow empty strings
        (
            "",
            dict(length=5, allow_empty=False),
            "Input is too short (length should be at least 5)",
        ),
    ],
)
def test_min_length(value, kwargs, error):
    validator = validators.MinLength(**kwargs)

    assert validator(value) == error


@pytest.mark.parametrize(
    "value, kwargs, error",
    [
        # Exact max size
        ("right size", dict(length=10), None),
        # Less than max size
        ("valid", dict(length=10), None),
        # Empty strings are less than any max length
        ("", dict(length=10), None),
        # Just exceeds maximum length
        (
            "bad size!!!",
            dict(length=10),
            "Input is too long (length should be at most 10)",
        ),
        # Exceeds maximum length
        (
            "this is a long string",
            dict(length=10),
            "Input is too long (length should be at most 10)",
        ),
        # Exceeds a different maximum length
        (
            "this is a long string",
            dict(length=5),
            "Input is too long (length should be at most 5)",
        ),
        # Custom error message
        (
            "this is a long string",
            dict(length=10, error_message="Badness"),
            "Badness",
        ),
    ],
)
def test_max_length(value, kwargs, error):
    validator = validators.MaxLength(**kwargs)

    assert validator(value) == error


@pytest.mark.parametrize(
    "value, kwargs, error",
    [
        # Exact max size
        ("right size", dict(min_length=5, max_length=10), None),
        # Middle size
        ("all good", dict(min_length=5, max_length=10), None),
        # Exact min size
        ("valid", dict(min_length=5, max_length=10), None),
        # Allow empty strings
        ("", dict(min_length=5, max_length=10), None),
        # Just Exceeds maximum length
        (
            "bad size!!!",
            dict(min_length=5, max_length=10),
            "Input should be between 5 and 10 characters",
        ),
        # Just less than minimum length
        (
            "bad!",
            dict(min_length=5, max_length=10),
            "Input should be between 5 and 10 characters",
        ),
        # Custom error message
        (
            "this is a long string",
            dict(min_length=5, max_length=10, error_message="Badness"),
            "Badness",
        ),
        # Don't allow empty strings
        (
            "",
            dict(min_length=5, max_length=10, allow_empty=False),
            "Input should be between 5 and 10 characters",
        ),
    ],
)
def test_length_between(value, kwargs, error):
    validator = validators.LengthBetween(**kwargs)

    assert validator(value) == error


def test_invalid_range():
    """Minimum value must be less than maximum value."""
    with pytest.raises(
        ValueError,
        match=r"Minimum length cannot be less than maximum length",
    ):
        validators.LengthBetween(min_length=10, max_length=5)


@pytest.mark.parametrize(
    "value, kwargs, error",
    [
        # Exact match
        ("hello", dict(substring="hello"), None),
        # Starts with match
        ("hello world", dict(substring="hello"), None),
        # Allow empty strings
        ("", dict(substring="hello"), None),
        # Doesn't start with match
        (
            "bad string",
            dict(substring="hello"),
            "Input should start with 'hello'",
        ),
        # Custom error message
        (
            "bad string",
            dict(substring="hello", error_message="Badness"),
            "Badness",
        ),
        # Don't allow empty strings
        (
            "",
            dict(substring="hello", allow_empty=False),
            "Input should start with 'hello'",
        ),
    ],
)
def test_startswith(value, kwargs, error):
    validator = validators.StartsWith(**kwargs)

    assert validator(value) == error


@pytest.mark.parametrize(
    "value, kwargs, error",
    [
        # Exact match
        ("goodbye", dict(substring="goodbye"), None),
        # Ends with match
        ("the final goodbye", dict(substring="goodbye"), None),
        # Allow empty strings
        ("", dict(substring="goodbye"), None),
        # Doesn't end with match
        (
            "bad string",
            dict(substring="goodbye"),
            "Input should end with 'goodbye'",
        ),
        # Custom error message
        (
            "bad string",
            dict(substring="goodbye", error_message="Badness"),
            "Badness",
        ),
        # Don't allow empty strings
        (
            "",
            dict(substring="goodbye", allow_empty=False),
            "Input should end with 'goodbye'",
        ),
    ],
)
def test_endswith(value, kwargs, error):
    validator = validators.EndsWith(**kwargs)

    assert validator(value) == error


@pytest.mark.parametrize(
    "value, kwargs, error",
    [
        # Exact single match
        ("hello", dict(substring="hello"), None),
        # Starts with match
        ("hello is what you should say", dict(substring="hello"), None),
        # Ends with match
        ("You should say hello", dict(substring="hello"), None),
        # Contains substring
        ("Say hello, you fool", dict(substring="hello"), None),
        # Contains multiple examples of substring
        ("Say hello, and hello again", dict(substring="hello"), None),
        # Contains exact match of multiple substrings
        ("Say hello, and hello again", dict(substring="hello", count=2), None),
        # Count of 0 validates non-existence
        ("Say hello, and hello again", dict(substring="bad", count=0), None),
        # Allow empty strings
        ("", dict(substring="hello"), None),
        # Doesn't contain match
        (
            "bad string",
            dict(substring="hello"),
            "Input should contain 'hello'",
        ),
        # Contain match with a count of 0
        (
            "hello world",
            dict(substring="hello", count=0),
            "Input should not contain 'hello'",
        ),
        # Contain match, but not the right count
        (
            "hello world",
            dict(substring="hello", count=2),
            "Input should contain 'hello' exactly 2 times",
        ),
        # Custom error message
        (
            "bad string",
            dict(substring="hello", error_message="Badness"),
            "Badness",
        ),
        # Don't allow empty strings
        (
            "",
            dict(substring="hello", allow_empty=False),
            "Input should contain 'hello'",
        ),
    ],
)
def test_contains(value, kwargs, error):
    validator = validators.Contains(**kwargs)

    assert validator(value) == error


@pytest.mark.parametrize(
    "value, kwargs, error",
    [
        # No match
        ("nothing to see", dict(substring="hello"), None),
        # Allow empty strings (an empty string can't contain a substring)
        ("", dict(substring="hello"), None),
        # Starts with match
        (
            "hello is what you should say",
            dict(substring="hello"),
            "Input should not contain 'hello'",
        ),
        # Ends with match
        (
            "You should say hello",
            dict(substring="hello"),
            "Input should not contain 'hello'",
        ),
        # Contains substring
        (
            "Say hello, you fool",
            dict(substring="hello"),
            "Input should not contain 'hello'",
        ),
        # Custom error message
        (
            "You should say hello",
            dict(substring="hello", error_message="Badness"),
            "Badness",
        ),
    ],
)
def test_not_contains(value, kwargs, error):
    validator = validators.NotContains(**kwargs)

    assert validator(value) == error


# Test regex matches 1 upper case, 2 lower case, then 1 upper case.
TEST_REGEX = r"[A-Z]{1}[a-z]{2}[A-Z]{1}"


@pytest.mark.parametrize(
    "value, kwargs, error",
    [
        # Exact match
        ("GooD", dict(regex_string=TEST_REGEX), None),
        # Match at start
        ("GooD is what this is", dict(regex_string=TEST_REGEX), None),
        # Match at end
        ("This is also GooD", dict(regex_string=TEST_REGEX), None),
        # Exact match in the middle of the string
        ("it's GooD if it's in the middle", dict(regex_string=TEST_REGEX), None),
        # Allow empty strings
        ("", dict(regex_string=TEST_REGEX), None),
        # Doesn't match
        (
            "no match here",
            dict(regex_string=TEST_REGEX),
            "Input should match regex: '[A-Z]{1}[a-z]{2}[A-Z]{1}'",
        ),
        # Custom error message
        (
            "no match here",
            dict(regex_string=TEST_REGEX, error_message="Badness"),
            "Badness",
        ),
        # Don't allow empty strings
        (
            "",
            dict(regex_string=TEST_REGEX, allow_empty=False),
            "Input should match regex: '[A-Z]{1}[a-z]{2}[A-Z]{1}'",
        ),
    ],
)
def test_regex(value, kwargs, error):
    validator = validators.MatchRegex(**kwargs)

    assert validator(value) == error


@pytest.mark.parametrize(
    "value, kwargs, error",
    [
        # Only uppercase
        ("GOOD!1", dict(), None),
        # Some uppercase
        ("Good!1", dict(), None),
        # Some uppercase
        ("GooD!1", dict(count=2), None),
        # Count of 0
        ("good!1", dict(count=0), None),
        # Allow empty strings
        ("", dict(), None),
        # Doesn't match
        (
            "no match here",
            dict(),
            "Input should contain at least one upper case character",
        ),
        # Bad count
        (
            "Good!1",
            dict(count=2),
            "Input should contain exactly 2 upper case characters",
        ),
        # Explicit count of 0
        (
            "Bad Text!1",
            dict(count=0),
            "Input should not contain upper case characters",
        ),
        # Custom error message
        (
            "no match here",
            dict(error_message="Badness"),
            "Badness",
        ),
        # Don't allow empty strings
        (
            "",
            dict(allow_empty=False),
            "Input should contain at least one upper case character",
        ),
    ],
)
def test_contains_uppercase(value, kwargs, error):
    validator = validators.ContainsUppercase(**kwargs)

    assert validator(value) == error


@pytest.mark.parametrize(
    "value, kwargs, error",
    [
        # Only lower case
        ("good!1", dict(), None),
        # Some lower case
        ("Good!1", dict(), None),
        # Some lower case, exact count
        ("GooD!1", dict(count=2), None),
        # Count of 0
        ("GOOD!1", dict(count=0), None),
        # Allow empty strings
        ("", dict(), None),
        # Doesn't match
        (
            "NO MATCH HERE!",
            dict(),
            "Input should contain at least one lower case character",
        ),
        # Bad count
        (
            "Good!1",
            dict(count=2),
            "Input should contain exactly 2 lower case characters",
        ),
        # Explicit count of 0
        (
            "Bad Text!1",
            dict(count=0),
            "Input should not contain lower case characters",
        ),
        # Custom error message
        (
            "NO MATCH HERE!",
            dict(error_message="Badness"),
            "Badness",
        ),
        # Don't allow empty strings
        (
            "",
            dict(allow_empty=False),
            "Input should contain at least one lower case character",
        ),
    ],
)
def test_contains_lowercase(value, kwargs, error):
    validator = validators.ContainsLowercase(**kwargs)

    assert validator(value) == error


@pytest.mark.parametrize(
    "value, kwargs, error",
    [
        # Only digits
        ("1234", dict(), None),
        # Some digits
        ("Good!12345", dict(), None),
        # Some digits, exact count
        ("GooD12!", dict(count=2), None),
        # Count of 0
        ("good!", dict(count=0), None),
        # Allow empty strings
        ("", dict(), None),
        # Doesn't match
        (
            "no match here",
            dict(),
            "Input should contain at least one digit",
        ),
        # Bad count
        (
            "Good!1",
            dict(count=2),
            "Input should contain exactly 2 digits",
        ),
        # Explicit count of 0
        (
            "Bad Text!1",
            dict(count=0),
            "Input should not contain digits",
        ),
        # Custom error message
        (
            "no match here",
            dict(error_message="Badness"),
            "Badness",
        ),
        # Don't allow empty strings
        (
            "",
            dict(allow_empty=False),
            "Input should contain at least one digit",
        ),
    ],
)
def test_contains_digit(value, kwargs, error):
    validator = validators.ContainsDigit(**kwargs)

    assert validator(value) == error


@pytest.mark.parametrize(
    "value, kwargs, error",
    [
        # Only special
        ("!@*&#^(*&!^(&@))", dict(), None),
        # Some special
        ("Good!1", dict(), None),
        # Some special, exact count
        ("GooD@!", dict(count=2), None),
        # Count of 0
        ("good1", dict(count=0), None),
        # Allow empty strings
        ("", dict(), None),
        # Doesn't match
        (
            "no match here",
            dict(),
            "Input should contain at least one special character",
        ),
        # Bad count
        (
            "Good!1",
            dict(count=2),
            "Input should contain exactly 2 special characters",
        ),
        # Explicit count of 0
        (
            "Bad Text!1",
            dict(count=0),
            "Input should not contain special characters",
        ),
        # Custom error message
        (
            "no match here",
            dict(error_message="Badness"),
            "Badness",
        ),
        # Don't allow empty strings
        (
            "",
            dict(allow_empty=False),
            "Input should contain at least one special character",
        ),
    ],
)
def test_contains_special(value, kwargs, error):
    validator = validators.ContainsSpecial(**kwargs)

    assert validator(value) == error


@pytest.mark.parametrize(
    "value, kwargs, error",
    [
        # Zero
        ("0", dict(), None),
        # positive integer
        ("123", dict(), None),
        # Negative integer
        ("-123", dict(), None),
        # Extra space
        (" 123 ", dict(), None),
        # leading zeros
        ("01234", dict(), None),
        # Allow empty strings
        ("", dict(), None),
        # Doesn't match
        (
            "no match here",
            dict(),
            "Input should be an integer",
        ),
        # Contains an integer, but not a pure integer
        (
            "no 123",
            dict(),
            "Input should be an integer",
        ),
        # Float
        (
            "1.234",
            dict(),
            "Input should be an integer",
        ),
        # Hex
        (
            "0x123",
            dict(),
            "Input should be an integer",
        ),
        # Octal
        (
            "0o123",
            dict(),
            "Input should be an integer",
        ),
        # Custom error message
        (
            "no match here",
            dict(error_message="Badness"),
            "Badness",
        ),
        # Don't allow empty strings
        (
            "",
            dict(allow_empty=False),
            "Input should be an integer",
        ),
    ],
)
def test_integer(value, kwargs, error):
    validator = validators.Integer(**kwargs)

    assert validator(value) == error


@pytest.mark.parametrize(
    "value, kwargs, error",
    [
        # Zero
        ("0", dict(), None),
        # positive integer
        ("123", dict(), None),
        # Negative integer
        ("-123", dict(), None),
        # Extra space
        (" 123 ", dict(), None),
        # leading zeros
        ("01234", dict(), None),
        # Float
        ("12.34", dict(), None),
        # Negative Float
        ("-12.34", dict(), None),
        # Float, no leading 0
        (".1234", dict(), None),
        # Negative Float, no leading 0
        ("-.1234", dict(), None),
        # Exponential
        ("1.23e+4", dict(), None),
        # Negative Exponential
        ("-1.23e-4", dict(), None),
        # Exponential (capitalized)
        ("1.23E+4", dict(), None),
        # Negative (capitalized)
        ("-1.23E-4", dict(), None),
        # Allow empty strings
        ("", dict(), None),
        # Doesn't match
        (
            "no match here",
            dict(),
            "Input should be a number",
        ),
        # Contains a number, but not a pure number
        (
            "no 1.23",
            dict(),
            "Input should be a number",
        ),
        # Just a decimal point isn't a number
        (
            ".",
            dict(),
            "Input should be a number",
        ),
        # Just an exponent isn't a number
        (
            "e+9",
            dict(),
            "Input should be a number",
        ),
        # Custom error message
        (
            "no match here",
            dict(error_message="Badness"),
            "Badness",
        ),
        # Don't allow empty strings
        (
            "",
            dict(allow_empty=False),
            "Input should be a number",
        ),
    ],
)
def test_number(value, kwargs, error):
    validator = validators.Number(**kwargs)

    assert validator(value) == error


@pytest.mark.parametrize(
    "value, kwargs, error",
    [
        # Valid email addresses
        ("tiberius@beeware.org", dict(), None),
        ("tiberius.yak@beeware.org", dict(), None),
        ("tiberius+yak@beeware.org", dict(), None),
        ("tiberius@beeware.ab.cd", dict(), None),
        ("tiberius@localhost", dict(), None),
        ("tiberius@beeware", dict(), None),
        ("2iberius@beeware.org", dict(), None),
        # Allow empty strings
        ("", dict(), None),
        # Invalid email addresses
        ("tiberius.beeware.org", dict(), "Input should be a valid email address"),
        ("tiberius@me@beeware.org", dict(), "Input should be a valid email address"),
        ("tiberius@beeware.", dict(), "Input should be a valid email address"),
        # Custom error message
        ("not an email", dict(error_message="badness"), "badness"),
        # Disallow empty strings
        ("", dict(allow_empty=False), "Input should be a valid email address"),
    ],
)
def test_email(value, kwargs, error):
    validator = validators.Email(**kwargs)

    assert validator(value) == error


# @pytest.mark.parametrize(
#     "value",
#     [
#     ],
# )
# def test_invalid_email(value):
#     validator = validators.Email()

#     assert validator(value) == "Input should be a valid email address"



from __future__ import annotations

from typing import TYPE_CHECKING

from toga.images import Image
from toga.platform import get_platform_factory

if TYPE_CHECKING:
    from toga.images import ImageT


class Screen:
    def __init__(self, _impl):
        self._impl = _impl
        self.factory = get_platform_factory()

    @property
    def name(self) -> str:
        """Unique name of the screen."""
        return self._impl.get_name()

    @property
    def origin(self) -> tuple[int, int]:
        """The absolute coordinates of the screen's origin, as a ``(x, y)`` tuple."""
        return self._impl.get_origin()

    @property
    def size(self) -> tuple[int, int]:
        """The size of the screen, as a ``(width, height)`` tuple."""
        return self._impl.get_size()

    def as_image(self, format: type[ImageT] = Image) -> ImageT:
        """Render the current contents of the screen as an image.

        :param format: Format to provide. Defaults to :class:`~toga.images.Image`; also
            supports :any:`PIL.Image.Image` if Pillow is installed, as well as any image
            types defined by installed :doc:`image format plugins
            </reference/plugins/image_formats>`.
        :returns: An image containing the screen content, in the format requested.
        """
        return Image(self._impl.get_image_data()).as_format(format)



import asyncio
import inspect
import sys
import traceback
import warnings
from abc import ABC


class NativeHandler:
    def __init__(self, handler):
        self.native = handler


async def long_running_task(interface, generator, cleanup):
    """Run a generator as an asynchronous coroutine."""
    try:
        try:
            while True:
                delay = next(generator)
                if delay:
                    await asyncio.sleep(delay)
        except StopIteration as e:
            result = e.value
    except Exception as e:
        print("Error in long running handler:", e, file=sys.stderr)
        traceback.print_exc()
    else:
        if cleanup:
            try:
                cleanup(interface, result)
            except Exception as e:
                print("Error in long running handler cleanup:", e, file=sys.stderr)
                traceback.print_exc()


async def handler_with_cleanup(handler, cleanup, interface, *args, **kwargs):
    try:
        result = await handler(interface, *args, **kwargs)
    except Exception as e:
        print("Error in async handler:", e, file=sys.stderr)
        traceback.print_exc()
    else:
        if cleanup:
            try:
                cleanup(interface, result)
            except Exception as e:
                print("Error in async handler cleanup:", e, file=sys.stderr)
                traceback.print_exc()


def wrapped_handler(interface, handler, cleanup=None):
    """Wrap a handler provided by the user, so it can be invoked.

    If the handler is a NativeHandler, return the handler object contained in the
    NativeHandler.

    If the handler is a bound method, or function, invoke it as it, and return the
    result.

    If the handler is a generator, invoke it asynchronously, with the yield values from
    the generator representing the duration to sleep between iterations.

    If the handler is a coroutine, install it on the asynchronous event loop.

    Returns either the native handler, or a wrapped function that will invoke the
    handler, using the interface as context. If a non-native handler, the wrapper
    function is annotated with the original handler function on the `_raw` attribute.
    """
    if handler:
        if isinstance(handler, NativeHandler):
            return handler.native

        def _handler(*args, **kwargs):
            if asyncio.iscoroutinefunction(handler):
                asyncio.ensure_future(
                    handler_with_cleanup(handler, cleanup, interface, *args, **kwargs)
                )
            else:
                try:
                    result = handler(interface, *args, **kwargs)
                except Exception as e:
                    print("Error in handler:", e, file=sys.stderr)
                    traceback.print_exc()
                else:
                    if inspect.isgenerator(result):
                        asyncio.ensure_future(
                            long_running_task(interface, result, cleanup)
                        )
                    else:
                        try:
                            if cleanup:
                                cleanup(interface, result)
                            return result
                        except Exception as e:
                            print("Error in handler cleanup:", e, file=sys.stderr)
                            traceback.print_exc()

        _handler._raw = handler

    else:
        # A dummy no-op handler
        def _handler(*args, **kwargs):
            try:
                if cleanup:
                    cleanup(interface, None)
            except Exception as e:
                print("Error in handler cleanup:", e, file=sys.stderr)
                traceback.print_exc()

        _handler._raw = None

    return _handler


class AsyncResult(ABC):
    def __init__(self, on_result=None):
        loop = asyncio.get_event_loop()
        self.future = loop.create_future()

        ######################################################################
        # 2023-12: Backwards compatibility
        ######################################################################
        if on_result:
            warnings.warn(
                "Synchronous `on_result` handlers have been deprecated; use `await` on the asynchronous result",
                DeprecationWarning,
            )

            self.on_result = on_result
        else:
            self.on_result = None
        ######################################################################
        # End backwards compatibility.
        ######################################################################

    def set_result(self, result):
        if not self.future.cancelled():
            self.future.set_result(result)
            if self.on_result:
                self.on_result(result)

    def set_exception(self, exc):
        if not self.future.cancelled():
            self.future.set_exception(exc)
            if self.on_result:
                self.on_result(None, exception=exc)

    def __repr__(self):
        return f"<Async {self.RESULT_TYPE} result; future={self.future}>"

    def __await__(self):
        return self.future.__await__()

    # All the comparison dunder methods are disabled
    def __bool__(self, other):
        raise RuntimeError(
            f"Can't check {self.RESULT_TYPE} result directly; use await or an on_result handler"
        )

    __lt__ = __bool__
    __le__ = __bool__
    __eq__ = __bool__
    __ne__ = __bool__
    __gt__ = __bool__
    __ge__ = __bool__


class PermissionResult(AsyncResult):
    RESULT_TYPE = "permission"



from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Any, Protocol, overload

import toga
from toga.handlers import wrapped_handler
from toga.platform import get_platform_factory

from .base import Widget

if TYPE_CHECKING:
    if sys.version_info < (3, 10):
        from typing_extensions import TypeAlias
    else:
        from typing import TypeAlias

    from toga.icons import IconContent

    OptionContainerContent: TypeAlias = (
        tuple[str, Widget]
        | tuple[str, Widget, IconContent | None]
        | tuple[str, Widget, IconContent | None, bool]
        | toga.OptionItem
    )


class OnSelectHandler(Protocol):
    def __call__(self, widget: OptionContainer, **kwargs: Any) -> None:
        """A handler that will be invoked when a new tab is selected in the OptionContainer.

        :param widget: The OptionContainer that had a selection change.
        :param kwargs: Ensures compatibility with arguments added in future versions.
        """
        ...


class OptionItem:
    def __init__(
        self,
        text: str,
        content: Widget,
        *,
        icon: IconContent | None = None,
        enabled: bool = True,
    ):
        """A tab of content in an OptionContainer.

        :param text: The text label for the new tab.
        :param content: The content widget to use for the new tab.
        :param icon: The :any:`icon content <IconContent>` to use to represent the tab.
        :param enabled: Should the new tab be enabled?
        """
        if content is None:
            raise ValueError("Content widget cannot be None.")

        self._content = content
        # These properties only exist while the item is in construction. Once the tab is
        # actual content, these attributes will be deleted and the native implementation
        # will become the source of truth. Initially prime the attributes with None (so
        # that the attribute exists), then use the setter to enforce validation on the
        # provided values.
        self._text = None
        self._icon = None
        self._enabled = None

        self.text = text
        self.icon = icon
        self.enabled = enabled

        # Prime the attributes for properties that will be set when the OptionItem is
        # set as content.
        self._interface = None
        self._index = None

    @property
    def interface(self) -> OptionContainer:
        """The OptionContainer that contains this tab.

        Returns ``None`` if the tab isn't currently part of an OptionContainer.
        """
        return self._interface

    @property
    def enabled(self) -> bool:
        """Is the panel of content available for selection?"""
        try:
            return self._enabled
        except AttributeError:
            return self._interface._impl.is_option_enabled(self.index)

    @enabled.setter
    def enabled(self, value):
        enable = bool(value)
        if hasattr(self, "_enabled"):
            self._enabled = enable
        else:
            if (
                not enable
                and self.index == self._interface._impl.get_current_tab_index()
            ):
                raise ValueError("The currently selected tab cannot be disabled.")

            self._interface._impl.set_option_enabled(self.index, enable)

    @property
    def text(self) -> str:
        """The label for the tab of content."""
        try:
            return self._text
        except AttributeError:
            return self._interface._impl.get_option_text(self.index)

    @text.setter
    def text(self, value):
        if value is None:
            raise ValueError("Item text cannot be None")

        text = str(value)
        if not text:
            raise ValueError("Item text cannot be blank")

        if hasattr(self, "_text"):
            self._text = text
        else:
            self._interface._impl.set_option_text(self.index, text)

    @property
    def icon(self) -> toga.Icon:
        """The Icon for the tab of content.

        Can be specified as any valid :any:`icon content <IconContent>`.

        If the platform does not support the display of icons, this property
        will return ``None`` regardless of any value provided.
        """
        try:
            return self._icon
        except AttributeError:
            return self._interface._impl.get_option_icon(self.index)

    @icon.setter
    def icon(self, icon_or_name: IconContent | None):
        if get_platform_factory().OptionContainer.uses_icons:
            if icon_or_name is None:
                icon = None
            elif isinstance(icon_or_name, toga.Icon):
                icon = icon_or_name
            else:
                icon = toga.Icon(icon_or_name)

            if hasattr(self, "_icon"):
                self._icon = icon
            else:
                self._interface._impl.set_option_icon(self.index, icon)

    @property
    def index(self) -> int | None:
        """The index of the tab in the OptionContainer.

        Returns ``None`` if the tab isn't currently part of an OptionContainer.
        """
        return self._index

    @property
    def content(self) -> Widget:
        """The content widget displayed in this tab of the OptionContainer."""
        return self._content

    def _preserve_option(self):
        # Move the ground truth back to the OptionItem instance
        self._text = self.text
        self._icon = self.icon
        self._enabled = self.enabled

        # Clear
        self._index = None
        self._interface = None

    def _add_as_option(self, index, interface):
        text = self._text
        del self._text

        icon = self._icon
        del self._icon

        enabled = self._enabled
        del self._enabled

        self._index = index
        self._interface = interface
        interface._impl.add_option(index, text, self.content._impl, icon)

        # The option now exists on the implementation; finalize the display properties
        # that can't be resolved until the implementation exists.
        interface.refresh()
        self.enabled = enabled


class OptionList:
    def __init__(self, interface):
        self.interface = interface
        self._options = []

    def __repr__(self):
        items = ", ".join(repr(option.text) for option in self)
        return f"<OptionList {items}>"

    def __getitem__(self, index: int | str | OptionItem) -> OptionItem:
        """Obtain a specific tab of content."""
        return self._options[self.index(index)]

    def __delitem__(self, index: int | str | OptionItem):
        """Same as :any:`remove`."""
        self.remove(index)

    def remove(self, index: int | str | OptionItem):
        """Remove the specified tab of content.

        The currently selected item cannot be deleted.

        :param index: The index where the new tab should be inserted.
        """
        index = self.index(index)
        if index == self.interface._impl.get_current_tab_index():
            raise ValueError("The currently selected tab cannot be deleted.")

        # Ensure that the current ground truth of the item to be deleted is preserved as
        # attributes on the item
        deleted_item = self._options[index]
        deleted_item._preserve_option()

        self.interface._impl.remove_option(index)
        del self._options[index]
        # Update the index for each of the options
        # after the one that was removed.
        for option in self._options[index:]:
            option._index -= 1

        # Refresh the widget
        self.interface.refresh()

    def __len__(self) -> int:
        """The number of tabs of content in the OptionContainer."""
        return len(self._options)

    def index(self, value: str | int | OptionItem):
        """Find the index of the tab that matches the given value.

        :param value: The value to look for. An integer is returned as-is;
            if an :any:`OptionItem` is provided, that item's index is returned;
            any other value will be converted into a string, and the first
            tab with a label matching that string will be returned.
        :raises ValueError: If no tab matching the value can be found.
        """
        if isinstance(value, int):
            return value
        elif isinstance(value, OptionItem):
            return value.index
        else:
            try:
                return next(filter(lambda item: item.text == str(value), self)).index
            except StopIteration:
                raise ValueError(f"No tab named {value!r}")

    @overload
    def append(
        self,
        text_or_item: OptionItem,
    ): ...

    @overload
    def append(
        self,
        text_or_item: str,
        content: Widget,
        *,
        icon: IconContent | None = None,
        enabled: bool = True,
    ): ...

    def append(
        self,
        text_or_item: str,
        content: Widget | None = None,
        *,
        icon: IconContent | None = None,
        enabled: bool = None,
    ):
        """Add a new tab of content to the OptionContainer.

        The new tab can be specified as an existing :any:`OptionItem` instance, or by
        specifying the full details of the new tab of content. If an :any:`OptionItem`
        is provided, specifying ``content``, ``icon`` or ``enabled`` will raise an
        error.

        :param text_or_item: An :any:`OptionItem`; or, the text label for the new tab.
        :param content: The content widget to use for the new tab.
        :param icon: The :any:`icon content <IconContent>` to use to represent the tab.
        :param enabled: Should the new tab be enabled? (Default: ``True``)
        """
        self.insert(len(self), text_or_item, content, icon=icon, enabled=enabled)

    @overload
    def insert(
        self,
        index: int | str | OptionItem,
        text_or_item: OptionItem,
    ): ...

    @overload
    def insert(
        self,
        index: int | str | OptionItem,
        text_or_item: str,
        content: Widget,
        *,
        icon: IconContent | None = None,
        enabled: bool = True,
    ): ...

    def insert(
        self,
        index: int | str | OptionItem,
        text_or_item: str | OptionItem,
        content: Widget | None = None,
        *,
        icon: IconContent | None = None,
        enabled: bool | None = None,
    ):
        """Insert a new tab of content to the OptionContainer at the specified index.

        The new tab can be specified as an existing :any:`OptionItem` instance, or by
        specifying the full details of the new tab of content. If an :any:`OptionItem`
        is provided, specifying ``content``, ``icon`` or ``enabled`` will raise an
        error.

        :param index: The index where the new tab should be inserted.
        :param text_or_item: An :any:`OptionItem`; or, the text label for the new tab.
        :param content: The content widget to use for the new tab.
        :param icon: The :any:`icon content <IconContent>` to use to represent the tab.
        :param enabled: Should the new tab be enabled? (Default: ``True``)
        """
        if isinstance(text_or_item, OptionItem):
            if content is not None:
                raise ValueError(
                    "Cannot specify content if using an OptionItem instance."
                )
            if icon is not None:
                raise ValueError("Cannot specify icon if using an OptionItem instance.")
            if enabled is not None:
                raise ValueError(
                    "Cannot specify enabled if using an OptionItem instance."
                )
            item = text_or_item
        else:
            # Create an interface wrapper for the option.
            item = OptionItem(
                text_or_item,
                content,
                icon=icon,
                enabled=enabled if enabled is not None else True,
            )

        # Convert the index into an integer, and assign to the item.
        index = self.index(index)

        # Add the option to the list maintained on the interface,
        # and increment the index of all items after the one that was added.
        self._options.insert(index, item)
        for option in self._options[index + 1 :]:
            option._index += 1

        # Add the content to the implementation.
        # This will cause the native implementation to be created.
        item._add_as_option(index, self.interface)


class OptionContainer(Widget):
    def __init__(
        self,
        id=None,
        style=None,
        content: list[OptionContainerContent] | None = None,
        on_select: toga.widgets.optioncontainer.OnSelectHandler | None = None,
    ):
        """Create a new OptionContainer.

        :param id: The ID for the widget.
        :param style: A style object. If no style is provided, a default style will be
            applied to the widget.
        :param content: The initial :any:`OptionContainer content
            <OptionContainerContent>` to display in the OptionContainer.
        :param on_select: Initial :any:`on_select` handler.
        """
        super().__init__(id=id, style=style)
        self._content = OptionList(self)
        self.on_select = None

        self._impl = self.factory.OptionContainer(interface=self)

        if content:
            for item in content:
                if isinstance(item, OptionItem):
                    self.content.append(item)
                else:
                    if len(item) == 2:
                        text, widget = item
                        icon = None
                        enabled = True
                    elif len(item) == 3:
                        text, widget, icon = item
                        enabled = True
                    elif len(item) == 4:
                        text, widget, icon, enabled = item
                    else:
                        raise ValueError(
                            "Content items must be an OptionItem instance, or "
                            "tuples of (title, widget), (title, widget, icon), or "
                            "(title, widget, icon, enabled)"
                        )

                    self.content.append(text, widget, enabled=enabled, icon=icon)

        self.on_select = on_select

    @property
    def enabled(self) -> bool:
        """Is the widget currently enabled? i.e., can the user interact with the widget?

        OptionContainer widgets cannot be disabled; this property will always return
        True; any attempt to modify it will be ignored.
        """
        return True

    @enabled.setter
    def enabled(self, value):
        pass

    def focus(self):
        """No-op; OptionContainer cannot accept input focus"""
        pass

    @property
    def content(self) -> OptionList:
        """The tabs of content currently managed by the OptionContainer."""
        return self._content

    @property
    def current_tab(self) -> OptionItem | None:
        """The currently selected tab of content, or ``None`` if there are no tabs,
        or the OptionContainer is in a state where no tab is currently selected.

        This property can also be set with an ``int`` index, or a ``str`` label.
        """
        index = self._impl.get_current_tab_index()
        if index is None:
            return None
        return self._content[index]

    @current_tab.setter
    def current_tab(self, value):
        index = self._content.index(value)
        if not self._impl.is_option_enabled(index):
            raise ValueError("A disabled tab cannot be made the current tab.")

        self._impl.set_current_tab_index(index)

    @Widget.app.setter
    def app(self, app):
        # Invoke the superclass property setter
        Widget.app.fset(self, app)

        # Also assign the app to the content in the container
        for item in self._content:
            item._content.app = app

    @Widget.window.setter
    def window(self, window):
        # Invoke the superclass property setter
        Widget.window.fset(self, window)

        # Also assign the window to the content in the container
        for item in self._content:
            item._content.window = window

    @property
    def on_select(self) -> toga.widgets.optioncontainer.OnSelectHandler:
        """The callback to invoke when a new tab of content is selected."""
        return self._on_select

    @on_select.setter
    def on_select(self, handler):
        self._on_select = wrapped_handler(self, handler)



from pathlib import Path
from unittest.mock import Mock

import pytest

import toga
from toga.command import Separator
from toga_dummy.utils import assert_action_performed_with


def assert_order(*items):
    for i in range(0, len(items) - 1):
        for j in range(i + 1, len(items)):
            assert items[i] < items[j]
            assert items[j] > items[i]

            # For good measure; check comparisons with other types
            assert not items[i] < None
            assert not items[i] < 42
            assert not items[i] > None
            assert not items[i] > 42


def test_separator(parent_group_1):
    """A separator can be created."""

    separator = Separator(group=parent_group_1)
    assert repr(separator) == "<Separator group=P1>"


def test_separator_eq(parent_group_1, parent_group_2):
    """Separator objects can be compared for equality."""

    separator_1a = Separator(parent_group_1)
    separator_1b = Separator(parent_group_1)
    separator_2 = Separator(parent_group_2)

    # Separators are equal to breaks in the same section, but not to other
    # sections
    assert separator_1a == separator_1a
    assert separator_1a == separator_1b
    assert separator_1a != separator_2

    # Separators aren't equal to non-separator objects
    assert separator_1a != 3


def test_create():
    """A command can be created with defaults."""
    cmd = toga.Command(None, "Test command")

    assert cmd.text == "Test command"
    assert cmd.shortcut is None
    assert cmd.tooltip is None
    assert cmd.group == toga.Group.COMMANDS
    assert cmd.section == 0
    assert cmd.order == 0
    assert cmd.action._raw is None

    assert (
        repr(cmd)
        == "<Command text='Test command' group=<Group text='Commands' order=30> section=0 order=0>"
    )


def test_change_action():
    """A command's action can be changed to another handler."""
    action1 = Mock()

    cmd = toga.Command(action1, "Test command")

    assert cmd.text == "Test command"
    assert cmd.shortcut is None
    assert cmd.tooltip is None
    assert cmd.group == toga.Group.COMMANDS
    assert cmd.section == 0
    assert cmd.order == 0
    assert cmd.action._raw == action1

    # Change the action to a something new
    action2 = Mock()
    cmd.action = action2

    assert cmd.action._raw == action2

    # Clear the action
    cmd.action = None
    assert cmd.action._raw is None


def test_create_explicit(app):
    """A command can be created with explicit arguments."""
    grp = toga.Group("Test group", order=10)

    handler = Mock()
    cmd = toga.Command(
        handler,
        text="Test command",
        tooltip="This is a test command",
        shortcut="t",
        group=grp,
        section=3,
        order=4,
    )

    assert cmd.text == "Test command"
    assert cmd.shortcut == "t"
    assert cmd.tooltip == "This is a test command"
    assert cmd.group == grp
    assert cmd.section == 3
    assert cmd.order == 4

    assert cmd.action._raw == handler

    assert (
        repr(cmd)
        == "<Command text='Test command' group=<Group text='Test group' order=10> section=3 order=4>"
    )


@pytest.mark.parametrize("construct", [True, False])
def test_icon_construction(app, construct):
    """The command icon can be set during construction."""
    if construct:
        icon = toga.Icon("path/to/icon")
    else:
        icon = "path/to/icon"

    cmd = toga.Command(None, "Test command", icon=icon)
    assert isinstance(cmd.icon, toga.Icon)
    assert cmd.icon.path == Path("path/to/icon")


@pytest.mark.parametrize("construct", [True, False])
def test_icon(app, construct):
    """The command icon can be changed."""
    if construct:
        icon = toga.Icon("path/to/icon")
    else:
        icon = "path/to/icon"

    cmd = toga.Command(None, "Test command")

    # No icon by default
    assert cmd.icon is None

    # Change icon
    cmd.icon = icon

    # Icon path matches
    assert isinstance(cmd.icon, toga.Icon)
    assert cmd.icon.path == Path("path/to/icon")


@pytest.mark.parametrize(
    "action, enabled, initial_state",
    [
        (Mock(), True, True),
        (Mock(), False, False),
        (None, True, False),
        (None, False, False),
    ],
)
def test_enable(action, enabled, initial_state):
    cmd = toga.Command(action, text="Test command", enabled=enabled)

    assert cmd.enabled is initial_state

    # Set enabled; triggers an implementation response
    cmd.enabled = True
    assert_action_performed_with(cmd, "set enabled", value=True)

    # Disable; triggers an implementation response
    cmd.enabled = False
    assert_action_performed_with(cmd, "set enabled", value=False)

    # Disable again; triggers an implementation response
    cmd.enabled = False
    assert_action_performed_with(cmd, "set enabled", value=False)

    # Set enabled; triggers an implementation response
    cmd.enabled = True
    assert_action_performed_with(cmd, "set enabled", value=True)


def test_order_by_text():
    """Commands are ordered by text when group, section and order match."""
    assert_order(
        toga.Command(None, "A"),
        toga.Command(None, "B"),
    )


def test_order_by_number():
    """Commands are ordered by number when group and section match."""
    assert_order(
        toga.Command(None, "B", order=1),
        toga.Command(None, "A", order=2),
    )


def test_order_by_section(parent_group_1):
    """Section ordering takes priority over order and text."""
    assert_order(
        toga.Command(None, "B", group=parent_group_1, section=1, order=2),
        toga.Command(None, "A", group=parent_group_1, section=2, order=1),
    )


def test_order_by_groups(parent_group_1, parent_group_2, child_group_1, child_group_2):
    """Commands are ordered by group over."""

    command_z = toga.Command(None, "Z", group=parent_group_1, order=1)
    command_y = toga.Command(None, "Y", group=child_group_1, order=1)
    command_x = toga.Command(None, "X", group=child_group_1, order=2)
    command_w = toga.Command(None, "W", group=child_group_1, order=4)
    command_b = toga.Command(None, "B", group=child_group_1, section=2, order=1)
    command_v = toga.Command(None, "V", group=parent_group_1, order=3)
    command_u = toga.Command(None, "U", group=child_group_2, order=1)
    command_t = toga.Command(None, "T", group=child_group_2, order=2)
    command_s = toga.Command(None, "S", group=parent_group_1, order=5)
    command_a = toga.Command(None, "A", group=parent_group_2, order=1)

    assert_order(
        command_z,
        command_y,
        command_x,
        command_w,
        command_b,
        command_v,
        command_u,
        command_t,
        command_s,
        command_a,
    )



from .accessors import to_accessor  # noqa: F401
from .base import Listener, Source  # noqa: F401
from .list_source import ListSource, Row  # noqa: F401
from .tree_source import Node, TreeSource  # noqa: F401
from .value_source import ValueSource  # noqa: F401



from travertino.constants import (  # noqa: F401
    BOLD,
    BOTTOM,
    CENTER,
    COLUMN,
    CURSIVE,
    FANTASY,
    HIDDEN,
    ITALIC,
    JUSTIFY,
    LEFT,
    LTR,
    MONOSPACE,
    NONE,
    NORMAL,
    OBLIQUE,
    RIGHT,
    ROW,
    RTL,
    SANS_SERIF,
    SERIF,
    SMALL_CAPS,
    SYSTEM,
    TOP,
    TRANSPARENT,
    VISIBLE,
)
from travertino.declaration import BaseStyle, Choices
from travertino.layout import BaseBox
from travertino.size import BaseIntrinsicSize

from toga.fonts import (
    FONT_STYLES,
    FONT_VARIANTS,
    FONT_WEIGHTS,
    SYSTEM_DEFAULT_FONT_SIZE,
    SYSTEM_DEFAULT_FONTS,
    Font,
)

######################################################################
# Display
######################################################################

PACK = "pack"

######################################################################
# Declaration choices
######################################################################

DISPLAY_CHOICES = Choices(PACK, NONE)
VISIBILITY_CHOICES = Choices(VISIBLE, HIDDEN)
DIRECTION_CHOICES = Choices(ROW, COLUMN)
ALIGNMENT_CHOICES = Choices(LEFT, RIGHT, TOP, BOTTOM, CENTER)

SIZE_CHOICES = Choices(NONE, integer=True)
FLEX_CHOICES = Choices(number=True)

PADDING_CHOICES = Choices(integer=True)

TEXT_ALIGN_CHOICES = Choices(LEFT, RIGHT, CENTER, JUSTIFY)
TEXT_DIRECTION_CHOICES = Choices(RTL, LTR)

COLOR_CHOICES = Choices(color=True)
BACKGROUND_COLOR_CHOICES = Choices(TRANSPARENT, color=True)

FONT_FAMILY_CHOICES = Choices(*SYSTEM_DEFAULT_FONTS, string=True)
FONT_STYLE_CHOICES = Choices(*FONT_STYLES)
FONT_VARIANT_CHOICES = Choices(*FONT_VARIANTS)
FONT_WEIGHT_CHOICES = Choices(*FONT_WEIGHTS)
FONT_SIZE_CHOICES = Choices(integer=True)


class Pack(BaseStyle):
    class Box(BaseBox):
        pass

    class IntrinsicSize(BaseIntrinsicSize):
        pass

    _depth = -1

    def _debug(self, *args):  # pragma: no cover
        print("    " * self.__class__._depth, *args)

    @property
    def _hidden(self):
        """Does this style declaration define an object that should be hidden"""
        return self.visibility == HIDDEN

    def apply(self, prop, value):
        if self._applicator:
            if prop == "text_align":
                if value is None:
                    if self.text_direction == RTL:
                        value = RIGHT
                    else:
                        value = LEFT
                self._applicator.set_text_alignment(value)
            elif prop == "text_direction":
                if self.text_align is None:
                    self._applicator.set_text_alignment(RIGHT if value == RTL else LEFT)
            elif prop == "color":
                self._applicator.set_color(value)
            elif prop == "background_color":
                self._applicator.set_background_color(value)
            elif prop == "visibility":
                self._applicator.set_hidden(value == HIDDEN)
            elif prop in (
                "font_family",
                "font_size",
                "font_style",
                "font_variant",
                "font_weight",
            ):
                self._applicator.set_font(
                    Font(
                        self.font_family,
                        self.font_size,
                        style=self.font_style,
                        variant=self.font_variant,
                        weight=self.font_weight,
                    )
                )
            else:
                # Any other style change will cause a change in layout geometry,
                # so perform a refresh.
                self._applicator.refresh()

    def layout(self, node, viewport):
        # self._debug("=" * 80)
        # self._debug(f"Layout root {node}, available {viewport.width}x{viewport.height}")
        self.__class__._depth = -1

        self._layout_node(
            node,
            alloc_width=viewport.width,
            alloc_height=viewport.height,
            use_all_height=True,  # root node uses all height
            use_all_width=True,  # root node uses all width
        )
        node.layout.content_top = node.style.padding_top
        node.layout.content_bottom = node.style.padding_bottom

        node.layout.content_left = node.style.padding_left
        node.layout.content_right = node.style.padding_right

    def _layout_node(
        self,
        node,
        alloc_width,
        alloc_height,
        use_all_width,
        use_all_height,
    ):
        self.__class__._depth += 1
        # self._debug(
        #     f"COMPUTE LAYOUT for {node} available "
        #     f"{alloc_width}{'+' if use_all_width else ''}"
        #     " x "
        #     f"{alloc_height}{'+' if use_all_height else ''}"
        # )

        # Establish available width
        if self.width != NONE:
            # If width is specified, use it
            available_width = self.width
            min_width = self.width
            # self._debug(f"SPECIFIED WIDTH {self.width}")
        else:
            # If no width is specified, assume we're going to use all
            # the available width. If there is an intrinsic width,
            # use it to make sure the width is at least the amount specified.
            available_width = max(
                0, (alloc_width - self.padding_left - self.padding_right)
            )
            # self._debug(f"INITIAL {available_width=}")
            if node.intrinsic.width is not None:
                # self._debug(f"INTRINSIC WIDTH {node.intrinsic.width}")
                try:
                    min_width = node.intrinsic.width.value
                    available_width = max(available_width, min_width)
                except AttributeError:
                    available_width = node.intrinsic.width
                    min_width = node.intrinsic.width

                # self._debug(f"ADJUSTED {available_width=}")
            else:
                # self._debug(f"AUTO {available_width=}")
                min_width = 0

        # Establish available height
        if self.height != NONE:
            # If height is specified, use it.
            available_height = self.height
            min_height = self.height
            # self._debug(f"SPECIFIED HEIGHT {self.height}")
        else:
            available_height = max(
                0,
                alloc_height - self.padding_top - self.padding_bottom,
            )
            # self._debug(f"INITIAL {available_height=}")
            if node.intrinsic.height is not None:
                # self._debug(f"INTRINSIC HEIGHT {node.intrinsic.height}")
                try:
                    min_height = node.intrinsic.height.value
                    available_height = max(available_height, min_height)
                except AttributeError:
                    available_height = node.intrinsic.height
                    min_height = node.intrinsic.height

                # self._debug(f"ADJUSTED {available_height=}")
            else:
                # self._debug(f"AUTO {available_height=}")
                min_height = 0

        if node.children:
            if self.direction == COLUMN:
                min_width, width, min_height, height = self._layout_column_children(
                    node,
                    available_width=available_width,
                    available_height=available_height,
                    use_all_height=use_all_height,
                    use_all_width=use_all_width,
                )
            else:
                min_width, width, min_height, height = self._layout_row_children(
                    node,
                    available_width=available_width,
                    available_height=available_height,
                    use_all_height=use_all_height,
                    use_all_width=use_all_width,
                )
            # self._debug(f"HAS CHILDREN {min_width=} {width=} {min_height=} {height=}")
        else:
            width = available_width
            height = available_height
            # self._debug(f"NO CHILDREN {min_width=} {width=} {min_height=} {height=}")

        # If an explicit width/height was given, that specification
        # overrides the width/height evaluated by the layout of children
        if self.width != NONE:
            width = self.width
            min_width = width
        if self.height != NONE:
            height = self.height
            min_height = height

        # self._debug(f"FINAL SIZE {min_width}x{min_height} {width}x{height}")
        node.layout.content_width = int(width)
        node.layout.content_height = int(height)

        node.layout.min_content_width = int(min_width)
        node.layout.min_content_height = int(min_height)

        # self._debug("END LAYOUT", node, node.layout)
        self.__class__._depth -= 1

    def _layout_row_children(
        self,
        node,
        available_width,
        available_height,
        use_all_width,
        use_all_height,
    ):
        # self._debug(f"LAYOUT ROW CHILDREN {available_width=} {available_height=}")
        # Pass 1: Lay out all children with a hard-specified width, or an
        # intrinsic non-flexible width. While iterating, collect the flex
        # total of remaining elements.
        flex_total = 0
        min_flex = 0
        width = 0
        min_width = 0
        remaining_width = available_width
        for child in node.children:
            # self._debug(f"PASS 1 {child}")
            if child.style.width != NONE:
                # self._debug(f"- fixed width {child.style.width}")
                child.style._layout_node(
                    child,
                    alloc_width=remaining_width,
                    alloc_height=available_height,
                    use_all_width=False,
                    use_all_height=child.style.direction == ROW,
                )
                child_content_width = child.layout.content_width
                # It doesn't matter how small the children can be laid out;
                # we have an intrinsic size; so don't use min_content_width
                min_child_content_width = child.layout.content_width
            elif child.intrinsic.width is not None:
                if hasattr(child.intrinsic.width, "value"):
                    if child.style.flex:
                        # self._debug(f"- intrinsic flex width {child.intrinsic.width}")
                        flex_total += child.style.flex
                        # Final child content size will be computed in pass 2, after the
                        # amount of flexible space is known. For now, set an initial
                        # content height based on the intrinsic size, which will be the
                        # minimum possible allocation.
                        child_content_width = child.intrinsic.width.value
                        min_child_content_width = child.intrinsic.width.value
                        min_flex += (
                            child.style.padding_left
                            + child.intrinsic.width.value
                            + child.style.padding_right
                        )
                    else:
                        # self._debug(f"- intrinsic non-flex {child.intrinsic.width=}")
                        child.style._layout_node(
                            child,
                            alloc_width=0,
                            alloc_height=available_height,
                            use_all_width=False,
                            use_all_height=child.style.direction == ROW,
                        )
                        child_content_width = child.layout.content_width
                        # It doesn't matter how small the children can be laid out;
                        # we have an intrinsic size; so don't use min_content_width
                        min_child_content_width = child.layout.content_width
                else:
                    # self._debug(f"- intrinsic {child.intrinsic.width=}")
                    child.style._layout_node(
                        child,
                        alloc_width=remaining_width,
                        alloc_height=available_height,
                        use_all_width=False,
                        use_all_height=child.style.direction == ROW,
                    )
                    child_content_width = child.layout.content_width
                    # It doesn't matter how small the children can be laid out;
                    # we have an intrinsic size; so don't use min_content_width
                    min_child_content_width = child.layout.content_width
            else:
                if child.style.flex:
                    # self._debug("- unspecified flex width")
                    flex_total += child.style.flex
                    # Final child content size will be computed in pass 2, after the
                    # amount of flexible space is known. For now, use 0 as the minimum,
                    # as that's the best hint the widget style can give.
                    child_content_width = 0
                    min_child_content_width = 0
                else:
                    # self._debug("- unspecified non-flex width")
                    child.style._layout_node(
                        child,
                        alloc_width=remaining_width,
                        alloc_height=available_height,
                        use_all_width=False,
                        use_all_height=child.style.direction == ROW,
                    )
                    child_content_width = child.layout.content_width
                    min_child_content_width = child.layout.min_content_width

            child_width = (
                child.style.padding_left
                + child_content_width
                + child.style.padding_right
            )
            width += child_width
            remaining_width -= child_width

            min_child_width = (
                child.style.padding_left
                + min_child_content_width
                + child.style.padding_right
            )
            min_width += min_child_width

            # self._debug(f"  {min_child_width=} {min_width=} {min_flex=}")
            # self._debug(f"  {child_width=} {width=} {remaining_width=}")

        if flex_total > 0:
            quantum = (remaining_width + min_flex) / flex_total
            # In an ideal flex layout, all flex children will have a width proportional
            # to their flex value. However, if a flex child has a flexible minimum width
            # constraint that is greater than the ideal width for a balanced flex layout,
            # they need to be removed from the flex calculation.
            # self._debug(f"PASS 1a; {quantum=}")
            for child in node.children:
                if child.style.flex and child.intrinsic.width is not None:
                    try:
                        ideal_width = quantum * child.style.flex
                        if child.intrinsic.width.value > ideal_width:
                            # self._debug(f"- {child} overflows ideal width")
                            flex_total -= child.style.flex
                            min_flex -= (
                                child.style.padding_left
                                + child.intrinsic.width.value
                                + child.style.padding_right
                            )
                    except AttributeError:
                        # Intrinsic width isn't flexible
                        pass

            if flex_total > 0:
                quantum = (remaining_width + min_flex) / flex_total
            else:
                quantum = 0
        else:
            quantum = 0
        # self._debug(f"END PASS 1; {min_width=} {width=} {min_flex=} {quantum=}")

        # Pass 2: Lay out children with an intrinsic flexible width,
        # or no width specification at all.
        for child in node.children:
            # self._debug(f"PASS 2 {child}")
            if child.style.width != NONE:
                # self._debug("- already laid out (explicit width)")
                pass
            elif child.style.flex:
                if child.intrinsic.width is not None:
                    try:
                        child_alloc_width = (
                            child.style.padding_left
                            + child.intrinsic.width.value
                            + child.style.padding_right
                        )
                        ideal_width = quantum * child.style.flex
                        # self._debug(f"- flexible intrinsic {child_alloc_width=}")
                        if ideal_width > child_alloc_width:
                            # self._debug(f"  {ideal_width=}")
                            child_alloc_width = ideal_width

                        child.style._layout_node(
                            child,
                            alloc_width=child_alloc_width,
                            alloc_height=available_height,
                            use_all_width=True,
                            use_all_height=child.style.direction == ROW,
                        )
                        # Our width calculation already takes into account the intrinsic
                        # width; that has now expanded as a result of layout, so adjust
                        # to use the new layout size. Min width may also change, by the
                        # same scheme, because the flex child can itself have children,
                        # and those grandchildren have now been laid out.
                        # self._debug(f"  sub {child.intrinsic.width.value=}")
                        # self._debug(f"  add {child.layout.content_width=}")
                        # self._debug(f"  add min {child.layout.min_content_width=}")
                        width = (
                            width
                            - child.intrinsic.width.value
                            + child.layout.content_width
                        )
                        min_width = (
                            min_width
                            - child.intrinsic.width.value
                            + child.layout.min_content_width
                        )
                    except AttributeError:
                        # self._debug("- already laid out (fixed intrinsic width)")
                        pass
                else:
                    if quantum:
                        # self._debug(f"- unspecified flex width with {quantum=}")
                        child_alloc_width = quantum * child.style.flex
                    else:
                        # self._debug("- unspecified flex width")
                        child_alloc_width = (
                            child.style.padding_left + child.style.padding_right
                        )

                    child.style._layout_node(
                        child,
                        alloc_width=child_alloc_width,
                        alloc_height=available_height,
                        use_all_width=True,
                        use_all_height=child.style.direction == ROW,
                    )
                    # We now know the final min_width/width that accounts for flexible
                    # sizing; add that to the overall.
                    # self._debug(f"  add {child.layout.min_content_width=}")
                    # self._debug(f"  add {child.layout.content_width=}")
                    width += child.layout.content_width
                    min_width += child.layout.min_content_width
            else:
                # self._debug("- already laid out (intrinsic non-flex width)")
                pass

            # self._debug(f"  {min_width=} {width=}")

        # self._debug(f"PASS 2 COMPLETE; USED {width=}")
        if use_all_width:
            width = max(width, available_width)
        # self._debug(f"COMPUTED {min_width=} {width=}")

        # Pass 3: Set the horizontal position of each child, and establish row height
        offset = 0
        height = 0
        min_height = 0
        for child in node.children:
            # self._debug(f"PASS 3: {child} AT HORIZONTAL {offset=}")
            if node.style.text_direction is RTL:
                # self._debug("- RTL")
                offset += child.layout.content_width + child.style.padding_right
                child.layout.content_left = width - offset
                offset += child.style.padding_left
            else:
                # self._debug("- LTR")
                offset += child.style.padding_left
                child.layout.content_left = offset
                offset += child.layout.content_width + child.style.padding_right

            child_height = (
                child.style.padding_top
                + child.layout.content_height
                + child.style.padding_bottom
            )
            height = max(height, child_height)

            min_child_height = (
                child.style.padding_top
                + child.layout.min_content_height
                + child.style.padding_bottom
            )
            min_height = max(min_height, min_child_height)

        # self._debug(f"ROW {min_height=} {height=}")
        if use_all_height:
            height = max(height, available_height)
        # self._debug(f"FINAL ROW {min_height=} {height=}")

        # Pass 4: set vertical position of each child.
        for child in node.children:
            # self._debug(f"PASS 4: {child}")
            extra = height - (
                child.layout.content_height
                + child.style.padding_top
                + child.style.padding_bottom
            )
            # self._debug(f"- row extra width {extra}")
            if self.alignment is BOTTOM:
                child.layout.content_top = extra + child.style.padding_top
                # self._debug(f"  align {child} to bottom {child.layout.content_top=}")
            elif self.alignment is CENTER:
                child.layout.content_top = int(extra / 2) + child.style.padding_top
                # self._debug(f"  align {child} to center {child.layout.content_top=}")
            else:
                child.layout.content_top = child.style.padding_top
                # self._debug(f"  align {child} to top {child.layout.content_top=}")

        return min_width, width, min_height, height

    def _layout_column_children(
        self,
        node,
        available_width,
        available_height,
        use_all_width,
        use_all_height,
    ):
        # self._debug(f"LAYOUT COLUMN CHILDREN {available_width=} {available_height=}")
        # Pass 1: Lay out all children with a hard-specified height, or an
        # intrinsic non-flexible height. While iterating, collect the flex
        # total of remaining elements.
        flex_total = 0
        min_flex = 0
        height = 0
        min_height = 0
        remaining_height = available_height
        for child in node.children:
            # self._debug(f"PASS 1 {child}")
            if child.style.height != NONE:
                # self._debug(f"- fixed height {child.style.height}")
                child.style._layout_node(
                    child,
                    alloc_width=available_width,
                    alloc_height=remaining_height,
                    use_all_width=child.style.direction == COLUMN,
                    use_all_height=False,
                )
                child_content_height = child.layout.content_height
                # It doesn't matter how small the children can be laid out;
                # we have an intrinsic size; so don't use min_content_height
                min_child_content_height = child.layout.content_height
            elif child.intrinsic.height is not None:
                if hasattr(child.intrinsic.height, "value"):
                    if child.style.flex:
                        # self._debug(f"- intrinsic flex height {child.intrinsic.height}")
                        flex_total += child.style.flex
                        # Final child content size will be computed in pass 2, after the
                        # amount of flexible space is known. For now, set an initial
                        # content height based on the intrinsic size, which will be the
                        # minimum possible allocation.
                        child_content_height = child.intrinsic.height.value
                        min_child_content_height = child.intrinsic.height.value
                        min_flex += (
                            child.style.padding_top
                            + child_content_height
                            + child.style.padding_bottom
                        )
                    else:
                        # self._debug(f"- intrinsic non-flex {child.intrinsic.height=}")
                        child.style._layout_node(
                            child,
                            alloc_width=available_width,
                            alloc_height=0,
                            use_all_width=child.style.direction == COLUMN,
                            use_all_height=False,
                        )
                        child_content_height = child.layout.content_height
                        # It doesn't matter how small the children can be laid out;
                        # we have an intrinsic size; so don't use min_content_height
                        min_child_content_height = child.layout.content_height
                else:
                    # self._debug(f"- intrinsic {child.intrinsic.height=}")
                    child.style._layout_node(
                        child,
                        alloc_width=available_width,
                        alloc_height=remaining_height,
                        use_all_width=child.style.direction == COLUMN,
                        use_all_height=False,
                    )
                    child_content_height = child.layout.content_height
                    # It doesn't matter how small the children can be laid out;
                    # we have an intrinsic size; so don't use min_content_height
                    min_child_content_height = child.layout.content_height
            else:
                if child.style.flex:
                    # self._debug("- unspecified flex height")
                    flex_total += child.style.flex
                    # Final child content size will be computed in pass 2, after the
                    # amount of flexible space is known. For now, use 0 as the minimum,
                    # as that's the best hint the widget style can give.
                    child_content_height = 0
                    min_child_content_height = 0
                else:
                    # self._debug("- unspecified non-flex height")
                    child.style._layout_node(
                        child,
                        alloc_width=available_width,
                        alloc_height=remaining_height,
                        use_all_width=child.style.direction == COLUMN,
                        use_all_height=False,
                    )
                    child_content_height = child.layout.content_height
                    min_child_content_height = child.layout.min_content_height

            child_height = (
                child.style.padding_top
                + child_content_height
                + child.style.padding_bottom
            )
            height += child_height
            remaining_height -= child_height

            min_child_height = (
                child.style.padding_top
                + min_child_content_height
                + child.style.padding_bottom
            )
            min_height += min_child_height

            # self._debug(f"  {min_child_height=} {min_height=} {min_flex=}")
            # self._debug(f"  {child_height=} {height=} {remaining_height=}")

        if flex_total > 0:
            quantum = (remaining_height + min_flex) / flex_total
            # In an ideal flex layout, all flex children will have a height proportional
            # to their flex value. However, if a flex child has a flexible minimum
            # height constraint that is greater than the ideal height for a balanced
            # flex layout, they need to be removed from the flex calculation.
            # self._debug(f"PASS 1a; {quantum=}")
            for child in node.children:
                if child.style.flex and child.intrinsic.height is not None:
                    try:
                        ideal_height = quantum * child.style.flex
                        if child.intrinsic.height.value > ideal_height:
                            # self._debug(f"- {child} overflows ideal height")
                            flex_total -= child.style.flex
                            min_flex -= (
                                child.style.padding_top
                                + child.intrinsic.height.value
                                + child.style.padding_bottom
                            )
                    except AttributeError:
                        # Intrinsic height isn't flexible
                        pass

            if flex_total > 0:
                quantum = (min_flex + remaining_height) / flex_total
            else:
                quantum = 0
        else:
            quantum = 0

        # self._debug(f"END PASS 1; {min_height=} {height=} {min_flex=} {quantum=}")

        # Pass 2: Lay out children with an intrinsic flexible height,
        # or no height specification at all.
        for child in node.children:
            # self._debug(f"PASS 2 {child}")
            if child.style.height != NONE:
                # self._debug("- already laid out (explicit height)")
                pass
            elif child.style.flex:
                if child.intrinsic.height is not None:
                    try:
                        child_alloc_height = (
                            child.style.padding_top
                            + child.intrinsic.height.value
                            + child.style.padding_bottom
                        )
                        ideal_height = quantum * child.style.flex
                        # self._debug(f"- flexible intrinsic {child_alloc_height=}")
                        if ideal_height > child_alloc_height:
                            # self._debug(f"  {ideal_height=}")
                            child_alloc_height = ideal_height

                        child.style._layout_node(
                            child,
                            alloc_width=available_width,
                            alloc_height=child_alloc_height,
                            use_all_width=child.style.direction == COLUMN,
                            use_all_height=True,
                        )
                        # Our height calculation already takes into account the
                        # intrinsic height; that has now expanded as a result of layout,
                        # so adjust to use the new layout size. Min height may also
                        # change, by the same scheme, because the flex child can itself
                        # have children, and those grandchildren have now been laid out.
                        # self._debug(f"  sub {child.intrinsic.height.value=}")
                        # self._debug(f"  add {child.layout.content_height}")
                        # self._debug(f"  add min {child.layout.min_content_height}")
                        height = (
                            height
                            - child.intrinsic.height.value
                            + child.layout.content_height
                        )
                        min_height = (
                            min_height
                            - child.intrinsic.height.value
                            + child.layout.min_content_height
                        )
                    except AttributeError:
                        # self._debug("- already laid out (fixed intrinsic height)")
                        pass
                else:
                    if quantum:
                        # self._debug(f"- unspecified flex height with {quantum=}")
                        child_alloc_height = quantum * child.style.flex
                    else:
                        # self._debug("- unspecified flex height")
                        child_alloc_height = (
                            child.style.padding_top + child.style.padding_bottom
                        )

                    child.style._layout_node(
                        child,
                        alloc_width=available_width,
                        alloc_height=child_alloc_height,
                        use_all_width=child.style.direction == COLUMN,
                        use_all_height=True,
                    )
                    # We now know the final min_height/height that accounts for flexible
                    # sizing; add that to the overall.
                    # self._debug(f"  add {child.layout.min_content_height=}")
                    # self._debug(f"  add {child.layout.content_height=}")
                    height += child.layout.content_height
                    min_height += child.layout.min_content_height

            else:
                # self._debug("- already laid out (intrinsic non-flex height)")
                pass

            # self._debug(f"  {min_height=} {height=}")

        # self._debug(f"PASS 2 COMPLETE; USED {height=}")
        if use_all_height:
            height = max(height, available_height)
        # self._debug(f"COMPUTED {min_height=} {height=}")

        # Pass 3: Set the vertical position of each element, and establish column width
        offset = 0
        width = 0
        min_width = 0
        for child in node.children:
            # self._debug(f"PASS 3: {child} AT VERTICAL OFFSET {offset}")
            offset += child.style.padding_top
            child.layout.content_top = offset
            offset += child.layout.content_height + child.style.padding_bottom
            child_width = (
                child.layout.content_width
                + child.style.padding_left
                + child.style.padding_right
            )
            width = max(width, child_width)

            min_child_width = (
                child.style.padding_left
                + child.layout.min_content_width
                + child.style.padding_right
            )
            min_width = max(min_width, min_child_width)

        # self._debug(f"ROW {min_width=} {width=}")
        if use_all_width:
            width = max(width, available_width)
        # self._debug(f"FINAL ROW {min_width=} {width=}")

        # Pass 4: set horizontal position of each child.
        for child in node.children:
            # self._debug(f"PASS 4: {child}")
            extra = width - (
                child.layout.content_width
                + child.style.padding_left
                + child.style.padding_right
            )
            # self._debug(f"-  row extra width {extra}")
            if self.alignment is RIGHT:
                child.layout.content_left = extra + child.style.padding_left
                # self._debug(f"  align {child} to right {child.layout.content_left=}")
            elif self.alignment is CENTER:
                child.layout.content_left = int(extra / 2) + child.style.padding_left
                # self._debug(f"  align {child} to center {child.layout.content_left=}")
            else:
                child.layout.content_left = child.style.padding_left
                # self._debug(f"  align {child} to left {child.layout.content_left=}")

        return min_width, width, min_height, height

    def __css__(self):
        css = []
        # display
        if self.display == NONE:
            css.append("display: none;")
        else:
            # if self.display != NONE, it must be pack; it will inherit
            # the pack definition from the Toga stylesheet.
            pass

        # visibility
        if self.visibility != VISIBLE:
            css.append(f"visibility: {self.visibility};")

        # direction
        css.append(f"flex-direction: {self.direction.lower()};")
        # flex
        if (self.width == NONE and self.direction == ROW) or (
            self.height == NONE and self.direction == COLUMN
        ):
            css.append(f"flex: {self.flex} 0 auto;")

        # width/flex
        if self.width != NONE:
            css.append(f"width: {self.width}px;")

        # height/flex
        if self.height != NONE:
            css.append(f"height: {self.height}px;")

        # alignment
        if self.direction == ROW:
            if self.alignment:
                if self.alignment == LEFT:
                    css.append("align-items: start;")
                elif self.alignment == RIGHT:
                    css.append("align-items: end;")
                elif self.alignment == CENTER:
                    css.append("align-items: center;")
        else:
            if self.alignment:
                if self.alignment == TOP:
                    css.append("align-items: start;")
                elif self.alignment == BOTTOM:
                    css.append("align-items: end;")
                elif self.alignment == CENTER:
                    css.append("align-items: center;")

        # padding_*
        if self.padding_top:
            css.append(f"margin-top: {self.padding_top}px;")
        if self.padding_bottom:
            css.append(f"margin-bottom: {self.padding_bottom}px;")
        if self.padding_left:
            css.append(f"margin-left: {self.padding_left}px;")
        if self.padding_right:
            css.append(f"margin-right: {self.padding_right}px;")

        # color
        if self.color:
            css.append(f"color: {self.color};")

        # background_color
        if self.background_color:
            css.append(f"background-color: {self.background_color};")

        # text_align
        if self.text_align:
            css.append(f"text-align: {self.text_align};")

        # text_direction
        if self.text_direction != LTR:
            css.append(f"text-direction: {self.text_direction};")

        # font-*
        if self.font_family != SYSTEM:
            if " " in self.font_family:
                css.append(f'font-family: "{self.font_family}";')
            else:
                css.append(f"font-family: {self.font_family};")
        if self.font_size != SYSTEM_DEFAULT_FONT_SIZE:
            css.append(f"font-size: {self.font_size}pt;")
        if self.font_weight != NORMAL:
            css.append(f"font-weight: {self.font_weight};")
        if self.font_style != NORMAL:
            css.append(f"font-style: {self.font_style};")
        if self.font_variant != NORMAL:
            css.append(f"font-variant: {self.font_variant};")

        return " ".join(css)


Pack.validated_property("display", choices=DISPLAY_CHOICES, initial=PACK)
Pack.validated_property("visibility", choices=VISIBILITY_CHOICES, initial=VISIBLE)
Pack.validated_property("direction", choices=DIRECTION_CHOICES, initial=ROW)
Pack.validated_property("alignment", choices=ALIGNMENT_CHOICES)

Pack.validated_property("width", choices=SIZE_CHOICES, initial=NONE)
Pack.validated_property("height", choices=SIZE_CHOICES, initial=NONE)
Pack.validated_property("flex", choices=FLEX_CHOICES, initial=0)

Pack.validated_property("padding_top", choices=PADDING_CHOICES, initial=0)
Pack.validated_property("padding_right", choices=PADDING_CHOICES, initial=0)
Pack.validated_property("padding_bottom", choices=PADDING_CHOICES, initial=0)
Pack.validated_property("padding_left", choices=PADDING_CHOICES, initial=0)
Pack.directional_property("padding%s")

Pack.validated_property("color", choices=COLOR_CHOICES)
Pack.validated_property("background_color", choices=BACKGROUND_COLOR_CHOICES)

Pack.validated_property("text_align", choices=TEXT_ALIGN_CHOICES)
Pack.validated_property("text_direction", choices=TEXT_DIRECTION_CHOICES, initial=LTR)

Pack.validated_property("font_family", choices=FONT_FAMILY_CHOICES, initial=SYSTEM)
# Pack.list_property('font_family', choices=FONT_FAMILY_CHOICES)
Pack.validated_property("font_style", choices=FONT_STYLE_CHOICES, initial=NORMAL)
Pack.validated_property("font_variant", choices=FONT_VARIANT_CHOICES, initial=NORMAL)
Pack.validated_property("font_weight", choices=FONT_WEIGHT_CHOICES, initial=NORMAL)
Pack.validated_property(
    "font_size", choices=FONT_SIZE_CHOICES, initial=SYSTEM_DEFAULT_FONT_SIZE
)
# Pack.composite_property([
#     'font_family', 'font_style', 'font_variant', 'font_weight', 'font_size'
#     FONT_CHOICES
# ])
